@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/src/RCS/configdata.c 2.126 2002/03/03 19:27:46 amb Exp $
+  $Header: /home/amb/wwwoffle/src/RCS/configdata.c 2.144 2003/10/19 13:27:08 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.7a.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.8.
   Configuration data functions.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1997,98,99,2000,01,02 Andrew M. Bishop
+  This file Copyright 1997,98,99,2000,01,02,03 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -15,19 +15,19 @@
 
 #include "autoconfig.h"
 
+#include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <ctype.h>
 
 #include <sys/types.h>
 #include <unistd.h>
 
-#include "wwwoffle.h"
+#include "io.h"
 #include "misc.h"
+#include "errors.h"
 #include "configpriv.h"
 #include "config.h"
 #include "sockets.h"
-#include "errors.h"
 
 
 #ifndef PATH_MAX
@@ -91,7 +91,7 @@ static ConfigSection startup_section={"StartUp",
 /* Options Section */
 
 /*+ The level of error logging +*/
-ConfigItem _LogLevel;
+ConfigItem LogLevel;
 
 /*+ The amount of time that a socket connection will wait for data. +*/
 ConfigItem SocketTimeout;
@@ -121,12 +121,18 @@ ConfigItem RunFetch;            /*+ fetch (start or stop). +*/
 /*+ The option to have lock files to stop some problems. +*/
 ConfigItem LockFiles;
 
-/*+ The option to reply with compressed content encoding. +*/
+/*+ The option to reply to the client with compressed content encoding. +*/
 ConfigItem ReplyCompressedData;
 
+/*+ The option to reply to the client with chunked transfer encoding. +*/
+ConfigItem ReplyChunkedData;
+
+/*+ The paths or file extensions that are allowed to be used for CGIs. +*/
+ConfigItem ExecCGI;
+
 /*+ The item definitions in the Options section. +*/
 static ConfigItemDef options_itemdefs[]={
- {"log-level"            ,&_LogLevel          ,0,0,Fixed,CfgLogLevel,"Important"},
+ {"log-level"            ,&LogLevel           ,0,0,Fixed,CfgLogLevel,"Important"},
  {"socket-timeout"       ,&SocketTimeout      ,0,0,Fixed,TimeSecs   ,"2m"       },
  {"dns-timeout"          ,&DNSTimeout         ,0,0,Fixed,TimeSecs   ,"1m"       },
  {"connect-timeout"      ,&ConnectTimeout     ,0,0,Fixed,TimeSecs   ,"30"       },
@@ -139,7 +145,9 @@ static ConfigItemDef options_itemdefs[]={
  {"run-autodial"         ,&RunAutodial        ,0,0,Fixed,PathName   ,NULL       },
  {"run-fetch"            ,&RunFetch           ,0,0,Fixed,PathName   ,NULL       },
  {"lock-files"           ,&LockFiles          ,0,0,Fixed,Boolean    ,"no"       },
- {"reply-compressed-data",&ReplyCompressedData,0,0,Fixed,Boolean    ,"no"       }
+ {"reply-compressed-data",&ReplyCompressedData,0,0,Fixed,Boolean    ,"no"       },
+ {"reply-chunked-data"   ,&ReplyChunkedData   ,0,0,Fixed,Boolean    ,"no"       },
+ {"exec-cgi"             ,&ExecCGI            ,0,1,Fixed,Url        ,NULL       }
 };
 
 /*+ Options section. +*/
@@ -150,6 +158,12 @@ static ConfigSection options_section={"Options",
 
 /* OnlineOptions section. */
 
+/*+ The option to allow or ignore the 'Pragma: no-cache' request when online. +*/
+ConfigItem PragmaNoCacheOnline;
+
+/*+ The option to allow or ignore the 'Cache-Control: no-cache' request when online. +*/
+ConfigItem CacheControlNoCacheOnline;
+
 /*+ The maximum age of a cached page to use in preference while online. +*/
 ConfigItem RequestChanged;
 
@@ -165,6 +179,12 @@ ConfigItem RequestNoCache;
 /*+ The option to re-request pages that have status code 302 (temporary redirection).+*/
 ConfigItem RequestRedirection;
 
+/*+ The option to re-request pages with a conditional request.+*/
+ConfigItem RequestConditional;
+
+/*+ The option to use Etags as a cache validator.+*/
+ConfigItem ValidateWithEtag;
+
 /*+ The option to try and get the requested URL without a password as well as with. +*/
 ConfigItem TryWithoutPassword;
 
@@ -178,22 +198,30 @@ ConfigItem IntrDownloadPercent;        /*+ more than a given percentage complete
 /*+ The option to keep downloads that time out. +*/
 ConfigItem TimeoutDownloadKeep;
 
-/*+ The option to request compressed content encoding. +*/
+/*+ The option to request from the server compressed content encoding. +*/
 ConfigItem RequestCompressedData;
 
+/*+ The option to request from the server chunked transfer encoding. +*/
+ConfigItem RequestChunkedData;
+
 /*+ The item definitions in the OnlineOptions section. +*/
 static ConfigItemDef onlineoptions_itemdefs[]={
- {"request-changed"        ,&RequestChanged       ,1,0,Fixed,TimeSecs  ,"10m"},
- {"request-changed-once"   ,&RequestChangedOnce   ,1,0,Fixed,Boolean   ,"yes"},
- {"request-expired"        ,&RequestExpired       ,1,0,Fixed,Boolean   ,"no" },
- {"request-no-cache"       ,&RequestNoCache       ,1,0,Fixed,Boolean   ,"no" },
- {"request-redirection"    ,&RequestRedirection   ,1,0,Fixed,Boolean   ,"no" },
- {"try-without-password"   ,&TryWithoutPassword   ,1,0,Fixed,Boolean   ,"yes"},
- {"intr-download-keep"     ,&IntrDownloadKeep     ,1,0,Fixed,Boolean   ,"no" },
- {"intr-download-size"     ,&IntrDownloadSize     ,1,0,Fixed,FileSize  ,"1"  },
- {"intr-download-percent"  ,&IntrDownloadPercent  ,1,0,Fixed,Percentage,"80" },
- {"timeout-download-keep"  ,&TimeoutDownloadKeep  ,1,0,Fixed,Boolean   ,"no" },
- {"request-compressed-data",&RequestCompressedData,1,0,Fixed,Boolean   ,"yes"}
+ {"pragma-no-cache"        ,&PragmaNoCacheOnline      ,1,0,Fixed,Boolean   ,"yes"},
+ {"cache-control-no-cache" ,&CacheControlNoCacheOnline,1,0,Fixed,Boolean   ,"yes"},
+ {"request-changed"        ,&RequestChanged           ,1,0,Fixed,TimeSecs  ,"10m"},
+ {"request-changed-once"   ,&RequestChangedOnce       ,1,0,Fixed,Boolean   ,"yes"},
+ {"request-expired"        ,&RequestExpired           ,1,0,Fixed,Boolean   ,"no" },
+ {"request-no-cache"       ,&RequestNoCache           ,1,0,Fixed,Boolean   ,"no" },
+ {"request-redirection"    ,&RequestRedirection       ,1,0,Fixed,Boolean   ,"no" },
+ {"request-conditional"    ,&RequestConditional       ,1,0,Fixed,Boolean   ,"yes"},
+ {"validate-with-etag"     ,&ValidateWithEtag         ,1,0,Fixed,Boolean   ,"yes"},
+ {"try-without-password"   ,&TryWithoutPassword       ,1,0,Fixed,Boolean   ,"yes"},
+ {"intr-download-keep"     ,&IntrDownloadKeep         ,1,0,Fixed,Boolean   ,"no" },
+ {"intr-download-size"     ,&IntrDownloadSize         ,1,0,Fixed,FileSize  ,"1"  },
+ {"intr-download-percent"  ,&IntrDownloadPercent      ,1,0,Fixed,Percentage,"80" },
+ {"timeout-download-keep"  ,&TimeoutDownloadKeep      ,1,0,Fixed,Boolean   ,"no" },
+ {"request-compressed-data",&RequestCompressedData    ,1,0,Fixed,Boolean   ,"yes"},
+ {"request-chunked-data"   ,&RequestChunkedData       ,1,0,Fixed,Boolean   ,"yes"}
 };
 
 /*+ OnlineOptions section. +*/
@@ -204,8 +232,11 @@ static ConfigSection onlineoptions_section={"OnlineOptions",
 
 /* OfflineOptions section. */
 
-/*+ The option to allow or ignore the 'Pragma: no-cache' request. +*/
-ConfigItem PragmaNoCache;
+/*+ The option to allow or ignore the 'Pragma: no-cache' request when offline. +*/
+ConfigItem PragmaNoCacheOffline;
+
+/*+ The option to allow or ignore the 'Cache-Control: no-cache' request when offline. +*/
+ConfigItem CacheControlNoCacheOffline;
 
 /*+ The option to not automatically make requests while offline but to need confirmation. +*/
 ConfigItem ConfirmRequests;
@@ -215,9 +246,10 @@ ConfigItem DontRequestOffline;
 
 /*+ The item definitions in the OfflineOptions section. +*/
 static ConfigItemDef offlineoptions_itemdefs[]={
- {"pragma-no-cache" ,&PragmaNoCache     ,1,0,Fixed,Boolean,"yes"},
- {"confirm-requests",&ConfirmRequests   ,1,0,Fixed,Boolean,"no" },
- {"dont-request"    ,&DontRequestOffline,1,0,Fixed,Boolean,"no" }
+ {"pragma-no-cache"       ,&PragmaNoCacheOffline      ,1,0,Fixed,Boolean,"yes"},
+ {"cache-control-no-cache",&CacheControlNoCacheOffline,1,0,Fixed,Boolean,"yes"},
+ {"confirm-requests"      ,&ConfirmRequests           ,1,0,Fixed,Boolean,"no" },
+ {"dont-request"          ,&DontRequestOffline        ,1,0,Fixed,Boolean,"no" }
 };
 
 /*+ OfflineOptions section. +*/
@@ -237,6 +269,12 @@ ConfigItem FetchImages;
 /*+ The option to also fetch webbug images. +*/
 ConfigItem FetchWebbugImages;
 
+/*+ The option to also fetch icon images (favourite/shortcut icons). +*/
+ConfigItem FetchIconImages;
+
+/*+ The option to only fetch images from the same host. +*/
+ConfigItem FetchSameHostImages;
+
 /*+ The option to also fetch frames. +*/
 ConfigItem FetchFrames;
 
@@ -248,12 +286,14 @@ ConfigItem FetchObjects;
 
 /*+ The item definitions in the FetchOptions section. +*/
 static ConfigItemDef fetchoptions_itemdefs[]={
- {"stylesheets"  ,&FetchStyleSheets ,1,0,Fixed,Boolean,"no" },
- {"images"       ,&FetchImages      ,1,0,Fixed,Boolean,"no" },
- {"webbug-images",&FetchWebbugImages,1,0,Fixed,Boolean,"yes"},
- {"frames"       ,&FetchFrames      ,1,0,Fixed,Boolean,"no" },
- {"scripts"      ,&FetchScripts     ,1,0,Fixed,Boolean,"no" },
- {"objects"      ,&FetchObjects     ,1,0,Fixed,Boolean,"no" }
+ {"stylesheets"          ,&FetchStyleSheets   ,1,0,Fixed,Boolean,"no" },
+ {"images"               ,&FetchImages        ,1,0,Fixed,Boolean,"no" },
+ {"webbug-images"        ,&FetchWebbugImages  ,1,0,Fixed,Boolean,"yes"},
+ {"icon-images"          ,&FetchIconImages    ,1,0,Fixed,Boolean,"no" },
+ {"only-same-host-images",&FetchSameHostImages,1,0,Fixed,Boolean,"no" },
+ {"frames"               ,&FetchFrames        ,1,0,Fixed,Boolean,"no" },
+ {"scripts"              ,&FetchScripts       ,1,0,Fixed,Boolean,"no" },
+ {"objects"              ,&FetchObjects       ,1,0,Fixed,Boolean,"no" }
 };
 
 /*+ The FetchOptions section. +*/
@@ -264,8 +304,8 @@ static ConfigSection fetchoptions_section={"FetchOptions",
 
 /* IndexOptions section */
 
-/*+ The option to disable the lasttime/prevtime indexes. +*/
-ConfigItem NoLasttimeIndex;
+/*+ The option to enable/disable the lasttime/prevtime/lastout/prevout indexes. +*/
+ConfigItem CreateHistoryIndexes;
 
 /*+ The option to cycle the last time directories daily. +*/
 ConfigItem CycleIndexesDaily;
@@ -287,13 +327,13 @@ ConfigItem IndexListAny;
 
 /*+ The item definitions in the IndexOptions section. +*/
 static ConfigItemDef indexoptions_itemdefs[]={
- {"no-lasttime-index"  ,&NoLasttimeIndex  ,0,0,Fixed,Boolean,"no" },
- {"cycle-indexes-daily",&CycleIndexesDaily,0,0,Fixed,Boolean,"no" },
- {"list-outgoing"      ,&IndexListOutgoing,1,0,Fixed,Boolean,"yes"},
- {"list-latest"        ,&IndexListLatest  ,1,0,Fixed,Boolean,"yes"},
- {"list-monitor"       ,&IndexListMonitor ,1,0,Fixed,Boolean,"yes"},
- {"list-host"          ,&IndexListHost    ,1,0,Fixed,Boolean,"yes"},
- {"list-any"           ,&IndexListAny     ,1,0,Fixed,Boolean,"yes"}
+ {"create-history-indexes",&CreateHistoryIndexes,0,0,Fixed,Boolean,"yes"},
+ {"cycle-indexes-daily"   ,&CycleIndexesDaily   ,0,0,Fixed,Boolean,"no" },
+ {"list-outgoing"         ,&IndexListOutgoing   ,1,0,Fixed,Boolean,"yes"},
+ {"list-latest"           ,&IndexListLatest     ,1,0,Fixed,Boolean,"yes"},
+ {"list-monitor"          ,&IndexListMonitor    ,1,0,Fixed,Boolean,"yes"},
+ {"list-host"             ,&IndexListHost       ,1,0,Fixed,Boolean,"yes"},
+ {"list-any"              ,&IndexListAny        ,1,0,Fixed,Boolean,"yes"}
 };
 
 /*+ The IndexOptions section. +*/
@@ -307,9 +347,6 @@ static ConfigSection indexoptions_section={"IndexOptions",
 /*+ The option to turn on the modifications in this section. +*/
 ConfigItem EnableHTMLModifications;
 
-/*+ The option to turn on the modifications when online. +*/
-ConfigItem EnableModificationsOnline;
-
 /*+ The option of a tag that can be added to the bottom of the spooled pages with the date and some buttons. +*/
 ConfigItem AddCacheInfo;
 
@@ -332,12 +369,15 @@ ConfigItem DisableHTMLBlink;
 /*+ The option to disable Shockwave Flash animations. +*/
 ConfigItem DisableHTMLFlash;
 
-/*+ The option to disable any <meta http-equiv=refresh content=""> tags. +*/
+/*+ The option to disable any <meta http-equiv=Refresh content=""> tags. +*/
 ConfigItem DisableHTMLMetaRefresh;
 
-/*+ The option to disable any <meta http-equiv=refresh content=""> tags that refer to the same URL. +*/
+/*+ The option to disable any <meta http-equiv=Refresh content=""> tags that refer to the same URL. +*/
 ConfigItem DisableHTMLMetaRefreshSelf;
 
+/*+ The option to disable any <meta http-equiv=Set-Cookie content=""> tags. +*/
+ConfigItem DisableHTMLMetaSetCookie;
+
 /*+ The option to disable links (anchors) to pages in the DontGet list. +*/
 ConfigItem DisableHTMLDontGetAnchors;
 
@@ -365,7 +405,6 @@ ConfigItem DisableAnimatedGIF;
 /*+ The item definitions in the ModifyHTMLOptions section. +*/
 static ConfigItemDef modifyhtml_itemdefs[]={
  {"enable-modify-html"       ,&EnableHTMLModifications    ,1,0,Fixed,Boolean,"no"},
- {"enable-modify-online"     ,&EnableModificationsOnline  ,1,0,Fixed,Boolean,"no"},
  {"add-cache-info"           ,&AddCacheInfo               ,1,0,Fixed,Boolean,"no"},
  {"anchor-cached-begin"      ,&AnchorModifyBegin[0]       ,1,0,Fixed,String ,NULL},
  {"anchor-cached-end"        ,&AnchorModifyEnd[0]         ,1,0,Fixed,String ,NULL},
@@ -380,6 +419,7 @@ static ConfigItemDef modifyhtml_itemdefs[]={
  {"disable-flash"            ,&DisableHTMLFlash           ,1,0,Fixed,Boolean,"no"},
  {"disable-meta-refresh"     ,&DisableHTMLMetaRefresh     ,1,0,Fixed,Boolean,"no"},
  {"disable-meta-refresh-self",&DisableHTMLMetaRefreshSelf ,1,0,Fixed,Boolean,"no"},
+ {"disable-meta-set-cookie"  ,&DisableHTMLMetaSetCookie   ,1,0,Fixed,Boolean,"no"},
  {"disable-dontget-links"    ,&DisableHTMLDontGetAnchors  ,1,0,Fixed,Boolean,"no"},
  {"disable-dontget-iframes"  ,&DisableHTMLDontGetIFrames  ,1,0,Fixed,Boolean,"no"},
  {"replace-dontget-images"   ,&ReplaceHTMLDontGetImages   ,1,0,Fixed,Boolean,"no"},
@@ -529,14 +569,18 @@ static ConfigSection dontcompress_section={"DontCompress",
 /*+ The list of censored headers. +*/
 ConfigItem CensorHeader;
 
-/*+ Flags to cause the referer header to be mangled. +*/
+/*+ Flags to cause the 'Referer' header to be mangled. +*/
 ConfigItem RefererSelf;
 ConfigItem RefererSelfDir;
 
+/*+ A flag to cause a 'User-Agent' header always to be added. +*/
+ConfigItem ForceUserAgent;
+
 /*+ The item definitions in the censor headers section. +*/
 static ConfigItemDef censorheader_itemdefs[]={
  {"referer-self"    ,&RefererSelf   ,1,0,Fixed ,Boolean,"no"},
  {"referer-self-dir",&RefererSelfDir,1,0,Fixed ,Boolean,"no"},
+ {"force-user-agent",&ForceUserAgent,1,0,Fixed ,Boolean,"no"},
  {""                ,&CensorHeader  ,1,1,String,String ,NULL}
 };
 
@@ -640,10 +684,10 @@ static ConfigSection alias_section={"Alias",
 ConfigItem PurgeUseMTime;
 
 /*+ The maximum allowed size of the cache. +*/
-ConfigItem PurgeCacheSize;
+ConfigItem PurgeMaxSize;
 
 /*+ The minimum allowed free disk space. +*/
-ConfigItem PurgeDiskFree;
+ConfigItem PurgeMinFree;
 
 /*+ A flag to indicate if the whole URL is used to choose the purge age. +*/
 ConfigItem PurgeUseURL;
@@ -665,8 +709,8 @@ static ConfigItemDef purge_itemdefs[]={
  {"use-mtime"    ,&PurgeUseMTime      ,0,0,Fixed,Boolean  ,"no"},
  {"age"          ,&PurgeAges          ,1,0,Fixed,AgeDays  ,"2w"},
  {"compress-age" ,&PurgeCompressAges  ,1,0,Fixed,AgeDays  ,"-1"},
- {"max-size"     ,&PurgeCacheSize     ,0,0,Fixed,CacheSize,"0" },
- {"min-free"     ,&PurgeDiskFree      ,0,0,Fixed,CacheSize,"0" },
+ {"max-size"     ,&PurgeMaxSize       ,0,0,Fixed,CacheSize,"-1"},
+ {"min-free"     ,&PurgeMinFree       ,0,0,Fixed,CacheSize,"-1"},
  {"use-url"      ,&PurgeUseURL        ,0,0,Fixed,Boolean  ,"no"},
  {"del-dontget"  ,&PurgeDontGet       ,0,0,Fixed,Boolean  ,"no"},
  {"del-dontcache",&PurgeDontCache     ,0,0,Fixed,Boolean  ,"no"}
@@ -760,11 +804,10 @@ void InitConfigurationFile(char *name)
 
  DefaultConfigFile();
 
- LogLevel=ConfigInteger(_LogLevel);
+ LoggingLevel=ConfigInteger(LogLevel);
 
  SetDNSTimeout(ConfigInteger(DNSTimeout));
  SetConnectTimeout(ConfigInteger(ConnectTimeout));
- set_read_timeout(ConfigInteger(SocketTimeout));
 }
 
 
@@ -790,11 +833,10 @@ int ReadConfigurationFile(int fd)
    }
  else
    {
-    LogLevel=ConfigInteger(_LogLevel);
+    LoggingLevel=ConfigInteger(LogLevel);
 
     SetDNSTimeout(ConfigInteger(DNSTimeout));
     SetConnectTimeout(ConfigInteger(ConnectTimeout));
-    set_read_timeout(ConfigInteger(SocketTimeout));
 
     return(0);
    }
