@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/src/RCS/configedit.c 1.15 2002/03/09 11:51:54 amb Exp $
+  $Header: /home/amb/wwwoffle/src/RCS/configedit.c 1.32 2003/10/19 13:27:08 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.7a.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.8.
   The HTML interactive configuration editing pages.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1997,98,99,2000,01,02 Andrew M. Bishop
+  This file Copyright 1997,98,99,2000,01,02,03 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -22,120 +22,19 @@
 #include <unistd.h>
 
 #include "wwwoffle.h"
+#include "io.h"
 #include "misc.h"
 #include "configpriv.h"
 #include "config.h"
-#include "errors.h"
 
 
-static void ConfigurationIndexPage(int fd,char *url);
-static void ConfigurationSectionPage(int fd,int section,char *url);
-static void ConfigurationItemPage(int fd,int section,int item,char *url,Body *request_body);
-static void ConfigurationURLPage(int fd,char *request_args);
+static void ConfigurationIndexPage(int fd,/*@null@*/ char *url);
+static void ConfigurationSectionPage(int fd,int section,/*@null@*/ char *url);
+static void ConfigurationItemPage(int fd,int section,int item,/*@null@*/ char *url,/*@null@*/ Body *request_body);
+static void ConfigurationEditURLPage(int fd,/*@null@*/ Body *request_body);
+static void ConfigurationURLPage(int fd,/*@null@*/ char *url);
 static void ConfigurationAuthFail(int fd,char *url);
 
-/*+ The list of configuration items to list on the configuration page for a URL. +*/
-static ConfigItem *URLConfigItems[]={
-
- /* OnlineOptions */
-
- &RequestChanged,
- &RequestChangedOnce,
- &RequestExpired,
- &RequestNoCache,
- &RequestRedirection,
- &TryWithoutPassword,
- &IntrDownloadKeep,
- &IntrDownloadSize,
- &IntrDownloadPercent,
- &TimeoutDownloadKeep,
- &RequestCompressedData,
-
- /* OfflineOptions */
-
- &PragmaNoCache,
- &ConfirmRequests,
- &DontRequestOffline,
-
- /* FetchOptions */
-
- &FetchStyleSheets,
- &FetchImages,
- &FetchWebbugImages,
- &FetchFrames,
- &FetchScripts,
- &FetchObjects,
-
- /* IndexOptions */
-
- &IndexListOutgoing,
- &IndexListLatest,
- &IndexListMonitor,
- &IndexListHost,
- &IndexListAny,
-
- /* ModifyHtml */
-
- &EnableHTMLModifications,
- &EnableModificationsOnline,
- &AddCacheInfo,
- &AnchorModifyBegin[0],
- &AnchorModifyBegin[1],
- &AnchorModifyBegin[2],
- &AnchorModifyEnd[0],
- &AnchorModifyEnd[1],
- &AnchorModifyEnd[2],
- &DisableHTMLScript,
- &DisableHTMLApplet,
- &DisableHTMLStyle,
- &DisableHTMLBlink,
- &DisableHTMLFlash,
- &DisableHTMLMetaRefresh,
- &DisableHTMLMetaRefreshSelf,
- &DisableHTMLDontGetAnchors,
- &DisableHTMLDontGetIFrames,
- &ReplaceHTMLDontGetImages,
- &ReplacementHTMLDontGetImage,
- &ReplaceHTMLWebbugImages,
- &ReplacementHTMLWebbugImage,
- &DemoroniseMSChars,
- &DisableAnimatedGIF,
-
- /* DontCache */
-
- &DontCache,
-
- /* DontGet */
-
- &DontGet,
- &DontGetReplacementURL,
- &DontGetRecursive,
- &DontGetLocation,
-
- /* CensorHeader */
-
- &CensorHeader,
- &RefererSelf,
- &RefererSelfDir,
-
- /* FTPOptions */
-
- &FTPAuthUser,
- &FTPAuthPass,
-
- /* Proxy */
-
- &Proxies,
- &ProxyAuthUser,
- &ProxyAuthPass,
- &SSLProxy,
-
- /* Purge */
-
- &PurgeAges,
- &PurgeCompressAges
-};
-
 
 /*++++++++++++++++++++++++++++++++++++++
   Send to the client one of the pages to configure WWWOFFLE using HTML.
@@ -144,12 +43,10 @@ static ConfigItem *URLConfigItems[]={
 
   URL *Url The Url that was requested.
 
-  Header *request_head The head of the HTTP request.
-
   Body *request_body The body of the HTTP request.
   ++++++++++++++++++++++++++++++++++++++*/
 
-void ConfigurationPage(int fd,URL *Url,Header *request_head,Body *request_body)
+void ConfigurationPage(int fd,URL *Url,Body *request_body)
 {
  int s,i;
  char *url=NULL;
@@ -170,30 +67,10 @@ void ConfigurationPage(int fd,URL *Url,Header *request_head,Body *request_body)
       }
    }
 
- /* Parse the arguments if any */
+ /* Extract the URL argument if any (ignoring all posted forms) */
 
- if(Url->args)
-   {
-    char **args,*pling;
-
-    if(*Url->args=='!' && (pling=strstr(Url->args,"!POST")))
-      {
-       *pling=0;
-       args=SplitFormArgs(Url->args+1);
-       *pling='!';
-      }
-    else
-       args=SplitFormArgs(Url->args);
-
-    for(i=0;args[i];i++)
-      {
-       if(!strncmp("url=",args[i],4) && args[i][4])
-          url=URLDecodeFormArgs(args[i]+4);
-      }
-
-    free(args[0]);
-    free(args);
-   }
+ if(Url->args && *Url->args!='!')
+    url=URLDecodeFormArgs(Url->args);
 
  /* Determine the page to show. */
 
@@ -204,9 +81,16 @@ void ConfigurationPage(int fd,URL *Url,Header *request_head,Body *request_body)
     return;
    }
 
+ if(!strcmp(&Url->path[15],"editurl"))
+   {
+    ConfigurationEditURLPage(fd,request_body);
+    if(url) free(url);
+    return;
+   }
+
  if(!strcmp(&Url->path[15],"url"))
    {
-    ConfigurationURLPage(fd,Url->args);
+    ConfigurationURLPage(fd,url);
     if(url) free(url);
     return;
    }
@@ -254,18 +138,20 @@ static void ConfigurationIndexPage(int fd,char *url)
 {
  char *line=NULL;
  int file;
+ off_t seekpos=0;
  int s;
 
- file=OpenLanguageFile("README.CONF.txt");
+ file=OpenLanguageFile("messages/README.CONF.txt");
 
  if(file==-1)
    {
-    HTMLMessage(fd,200,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
+    HTMLMessage(fd,500,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
                 "error","Cannot open README.CONF.txt",
                 NULL);
     return;
    }
 
+ init_io(file);
 
  line=read_line(file,line); /* TITLE ... */
  line=read_line(file,line); /* HEAD */
@@ -282,7 +168,7 @@ static void ConfigurationIndexPage(int fd,char *url)
     char *description=NULL;
 
     lseek(file,0,SEEK_SET);
-    init_buffer(file);
+    reinit_io(file);
 
     while((line=read_line(file,line)))
       {
@@ -294,6 +180,11 @@ static void ConfigurationIndexPage(int fd,char *url)
           if(line)
              line[strlen(line)-1]=0;
           description=line;
+          seekpos=lseek(file,0,SEEK_CUR);
+          if(seekpos>READ_BUFFER_SIZE)
+             seekpos-=READ_BUFFER_SIZE;
+          else
+             seekpos=0;
           break;
          }
       }
@@ -305,7 +196,8 @@ static void ConfigurationIndexPage(int fd,char *url)
                     NULL);
    }
 
- lseek(file,0,SEEK_SET);
+ lseek(file,seekpos,SEEK_SET);  /* go back only to the start of the last section */
+ reinit_io(file);
 
  while((line=read_line(file,line)))
    {
@@ -318,6 +210,12 @@ static void ConfigurationIndexPage(int fd,char *url)
  HTMLMessageBody(fd,"ConfigurationPage-Tail",
                  "description",line,
                  NULL);
+
+ if(line)
+    free(line);
+
+ finish_io(file);
+ close(file);
 }
 
 
@@ -338,16 +236,17 @@ static void ConfigurationSectionPage(int fd,int section,char *url)
  int file;
  off_t seekpos=0;
 
- file=OpenLanguageFile("README.CONF.txt");
+ file=OpenLanguageFile("messages/README.CONF.txt");
 
  if(file==-1)
    {
-    HTMLMessage(fd,200,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
+    HTMLMessage(fd,500,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
                 "error","Cannot open README.CONF.txt",
                 NULL);
     return;
    }
 
+ init_io(file);
 
  HTMLMessageHead(fd,200,"WWWOFFLE Configuration Section Page",
                  NULL);
@@ -388,8 +287,8 @@ static void ConfigurationSectionPage(int fd,int section,char *url)
     char *template=NULL;
     description=NULL;
 
-    lseek(file,seekpos,SEEK_SET);
-    init_buffer(file);
+    lseek(file,seekpos,SEEK_SET);  /* go back only to the start of the required section */
+    reinit_io(file);
 
     while((line1=read_line(file,line1)))
       {
@@ -403,6 +302,8 @@ static void ConfigurationSectionPage(int fd,int section,char *url)
       {
        line1[strlen(line1)-1]=0;
 
+       if(!strncmp(line1,"SECTION",7))
+          break;
        if(!strncmp(line1,"ITEM",4) && !strcmp(line1+5,CurrentConfig.sections[section]->itemdefs[i].name))
          {
           line1=read_line(file,line1);
@@ -431,6 +332,9 @@ static void ConfigurationSectionPage(int fd,int section,char *url)
     free(line1);
  if(line2)
     free(line2);
+
+ finish_io(file);
+ close(file);
 }
 
 
@@ -450,7 +354,7 @@ static void ConfigurationSectionPage(int fd,int section,char *url)
 
 static void ConfigurationItemPage(int fd,int section,int item,char *url,Body *request_body)
 {
- char *edit=NULL,*action=NULL,*entry=NULL;
+ char *action=NULL,*entry=NULL;
  char *key=NULL,*val=NULL;
  int i;
 
@@ -464,15 +368,15 @@ static void ConfigurationItemPage(int fd,int section,int item,char *url,Body *re
     for(i=0;args[i];i++)
       {
        if(!strncmp("url=",args[i],4) && args[i][4])
-          url=URLDecodeFormArgs(args[i]+4);
+          url=TrimArgs(URLDecodeFormArgs(args[i]+4));
        if(!strncmp("key=",args[i],4) && args[i][4])
-          key=URLDecodeFormArgs(args[i]+4);
+          key=TrimArgs(URLDecodeFormArgs(args[i]+4));
        if(!strncmp("val=",args[i],4) && args[i][4])
-          val=URLDecodeFormArgs(args[i]+4);
+          val=TrimArgs(URLDecodeFormArgs(args[i]+4));
        if(!strncmp("action=",args[i],7))
-          action=URLDecodeFormArgs(args[i]+7);
+          action=TrimArgs(URLDecodeFormArgs(args[i]+7));
        if(!strncmp("entry=",args[i],6))
-          entry=URLDecodeFormArgs(args[i]+6);
+          entry=TrimArgs(URLDecodeFormArgs(args[i]+6));
       }
 
     free(args[0]);
@@ -487,33 +391,43 @@ static void ConfigurationItemPage(int fd,int section,int item,char *url,Body *re
     char *line1=NULL,*line2=NULL,*template=NULL,*description=NULL;
     int file;
     char nentries[8],nallowed[8];
-    char *edit_url=NULL,*edit_key=NULL,*edit_val=NULL,*edit=NULL;
+    char *entry_url=NULL,*entry_key=NULL,*entry_val=NULL;
+    int edit=0;
 
-    if(action && !strcmp(action,"edit"))
+    if(action && entry)
       {
+       edit=1;
+
        for(i=0;i<(*CurrentConfig.sections[section]->itemdefs[item].item)->nentries;i++)
          {
-          edit=ConfigEntryString(*CurrentConfig.sections[section]->itemdefs[item].item,i);
+          char *thisentry=ConfigEntryString(*CurrentConfig.sections[section]->itemdefs[item].item,i);
 
-          if(!strcmp(edit,entry))
+          if(!strcmp(thisentry,entry))
+            {
+             ConfigEntryStrings(*CurrentConfig.sections[section]->itemdefs[item].item,i,&entry_url,&entry_key,&entry_val);
+             free(thisentry);
              break;
+            }
 
-          free(edit);
+          free(thisentry);
          }
       }
 
     /* Display the HTML */
 
-    file=OpenLanguageFile("README.CONF.txt");
+    file=OpenLanguageFile("messages/README.CONF.txt");
 
     if(file==-1)
       {
-       HTMLMessage(fd,200,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
+       HTMLMessage(fd,500,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
                    "error","Cannot open README.CONF.txt",
                    NULL);
+
+       if(entry) free(entry);
        return;
       }
 
+    init_io(file);
 
     HTMLMessageHead(fd,200,"WWWOFFLE Configuration Item Page",
                     NULL);
@@ -577,11 +491,11 @@ static void ConfigurationItemPage(int fd,int section,int item,char *url,Body *re
                               CurrentConfig.sections[section]->itemdefs[item].name:"",
                     "def_val",CurrentConfig.sections[section]->itemdefs[item].def_val,
 
-                    "url",edit?edit_url:url,
-                    "key",edit?edit_key:(!CurrentConfig.sections[section]->itemdefs[item].url_type &&
-                                         CurrentConfig.sections[section]->itemdefs[item].key_type==UrlSpecification &&
-                                         CurrentConfig.sections[section]->itemdefs[item].val_type==None)?url:key,
-                    "val",edit?edit_val:val,
+                    "url",edit?entry_url:url,
+                    "key",edit?entry_key:(!CurrentConfig.sections[section]->itemdefs[item].url_type &&
+                                          CurrentConfig.sections[section]->itemdefs[item].key_type==UrlSpecification &&
+                                          CurrentConfig.sections[section]->itemdefs[item].val_type==None)?url:key,
+                    "val",edit?entry_val:val,
 
                     "entry",entry,
 
@@ -629,18 +543,17 @@ static void ConfigurationItemPage(int fd,int section,int item,char *url,Body *re
     HTMLMessageBody(fd,"ConfigurationItem-Tail",
                     NULL);
 
-    if(edit)
-      {
-       if(edit_url) free(edit_url);
-       if(edit_key) free(edit_key);
-       if(edit_val) free(edit_val);
-       free(edit);
-      }
+    if(entry_url) free(entry_url);
+    if(entry_key) free(entry_key);
+    if(entry_val) free(entry_val);
 
     if(line1)
        free(line1);
     if(line2)
        free(line2);
+
+    finish_io(file);
+    close(file);
    }
 
     /* Display the page with the results of editing the parameters. */
@@ -687,187 +600,228 @@ static void ConfigurationItemPage(int fd,int section,int item,char *url,Body *re
 
  if(key)    free(key);
  if(val)    free(val);
- if(edit)   free(edit);
  if(action) free(action);
  if(entry)  free(entry);
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  The configuration page to edit the URL-SPECIFICATION and redirect to the real page.
+
+  int fd The file descriptor to write to.
+
+  Body *request_body The body of the request.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static void ConfigurationEditURLPage(int fd,Body *request_body)
+{
+ char *url=NULL,*encurl;
+ int urllen;
+ char *proto=NULL,*host=NULL,*port=NULL,*path=NULL,*args=NULL;
+ char *proto_other=NULL,*host_other=NULL,*port_other=NULL,*path_other=NULL,*args_other=NULL;
+ char *localhost,*relocate;
+
+ /* Parse the arguments. */
+
+ if(request_body)
+   {
+    char *form=URLRecodeFormArgs(request_body->content);
+    char **arglist=SplitFormArgs(form);
+    int i;
+
+    for(i=0;arglist[i];i++)
+      {
+       if(!strncmp("proto=",arglist[i],6) && arglist[i][6])
+          proto=TrimArgs(URLDecodeFormArgs(arglist[i]+6));
+       if(!strncmp("host=",arglist[i],5) && arglist[i][5])
+          host=TrimArgs(URLDecodeFormArgs(arglist[i]+5));
+       if(!strncmp("port=",arglist[i],5) && arglist[i][5])
+          port=TrimArgs(URLDecodeFormArgs(arglist[i]+5));
+       if(!strncmp("path=",arglist[i],5) && arglist[i][5])
+          path=TrimArgs(URLDecodeFormArgs(arglist[i]+5));
+       if(!strncmp("args=",arglist[i],5) && arglist[i][5])
+          args=TrimArgs(URLDecodeFormArgs(arglist[i]+5));
+
+       if(!strncmp("proto_other=",arglist[i],12) && arglist[i][12])
+          proto_other=TrimArgs(URLDecodeFormArgs(arglist[i]+12));
+       if(!strncmp("host_other=",arglist[i],11) && arglist[i][11])
+          host_other=TrimArgs(URLDecodeFormArgs(arglist[i]+11));
+       if(!strncmp("port_other=",arglist[i],11) && arglist[i][11])
+          port_other=TrimArgs(URLDecodeFormArgs(arglist[i]+11));
+       if(!strncmp("path_other=",arglist[i],11) && arglist[i][11])
+          path_other=TrimArgs(URLDecodeFormArgs(arglist[i]+11));
+       if(!strncmp("args_other=",arglist[i],11) && arglist[i][11])
+          args_other=TrimArgs(URLDecodeFormArgs(arglist[i]+11));
+      }
+
+    free(arglist[0]);
+    free(arglist);
+    free(form);
+   }
+
+ /* Sort out a URL from the mess of arguments. */
+
+ if(proto && !strcmp(proto,"OTHER") && proto_other)
+   {free(proto); proto=proto_other;}
+ else
+   {if(proto_other) free(proto_other);}
+
+ if(host && !strcmp(host,"OTHER") && host_other)
+   {free(host); host=host_other;}
+ else
+   {if(host_other) free(host_other);}
+
+ if(port && !strcmp(port,"OTHER") && port_other)
+   {free(port); port=(char*)malloc(strlen(port_other)+2); strcpy(port,":"); strcat(port,port_other);}
+ else
+   {if(port_other) free(port_other);}
+
+ if(path && !strcmp(path,"OTHER") && path_other)
+   {free(path); path=path_other;}
+ else
+   {if(path_other) free(path_other);}
+
+ if(args && !strcmp(args,"OTHER") && args_other)
+   {free(args); args=(char*)malloc(strlen(args_other)+2); strcpy(args,"?"); strcat(args,args_other);}
+ else
+   {if(args_other) free(args_other);}
+
+ if(!proto)
+    proto=strcpy(malloc(2),"*");
+
+ if(!host)
+    host=strcpy(malloc(2),"*");
+
+ if(!path)
+    path=strcpy(malloc(2),"*");
+
+ /* Create the URL */
+
+ urllen=4;
+ urllen+=strlen(proto);
+ urllen+=strlen(host);
+ if(port)
+    urllen+=strlen(port);
+ urllen+=strlen(path);
+ if(args)
+    urllen+=strlen(args);
+
+ url=(char*)malloc(urllen);
+
+ strcpy(url,proto);
+ strcat(url,"://");
+ strcat(url,host);
+ if(port)
+    strcat(url,port);
+ strcat(url,path);
+ if(args)
+    strcat(url,args);
+
+ if(proto)free(proto);
+ if(host) free(host);
+ if(port) free(port);
+ if(path) free(path);
+ if(args) free(args);
+
+ /* Redirect the client to it */
+
+ localhost=GetLocalHost(1);
+ encurl=URLEncodeFormArgs(url);
+ relocate=(char*)malloc(strlen(localhost)+strlen(encurl)+32);
+
+ sprintf(relocate,"http://%s/configuration/url?%s",localhost,encurl);
+
+ HTMLMessage(fd,302,"WWWOFFLE Configuration Edit URL Redirect",relocate,"Redirect",
+             "location",relocate,
+             NULL);
+
+ free(url);
+ free(encurl);
+ free(localhost);
+ free(relocate);
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   The configuration page for a specific URL.
 
   int fd The file descriptor to write to.
 
-  char *request_args The arguments to the configuration page.
+  char *url The URL specification from the URL argument.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static void ConfigurationURLPage(int fd,char *request_args)
+static void ConfigurationURLPage(int fd,char *url)
 {
- char *url=NULL;
- URL *Url=NULL;
  char *proto=NULL,*host=NULL,*port=NULL,*path=NULL,*args=NULL;
- char *proto_other=NULL,*host_other=NULL,*port_other=NULL,*path_other=NULL,*args_other=NULL;
- int wildcard=0;
- int file;
+ char *copy,*colon,*slash,*ques;
+ int wildcard=0,file,s,i,seekpos=0;
  char *line1=NULL,*line2=NULL;
- int configitem,s,i,last_s=-1,seekpos=0;
- char *form,**arglist;
 
- /* Parse the arguments. */
+ /* Assume a "well-formed" URL, from the function above or a WWWOFFLE index. */
 
- form=URLRecodeFormArgs(request_args);
- arglist=SplitFormArgs(form);
+ copy=(char*)malloc(strlen(url)+1);
+ strcpy(copy,url);
 
- for(i=0;arglist[i];i++)
+ colon=strchr(copy,':');
+ slash=strchr(copy,'/');
+
+ proto=copy;
+
+ host=slash+2;
+
+ if(colon && slash && slash==colon+1 && *(slash+1)=='/')
    {
-    if(!strncmp("url=",arglist[i],4) && arglist[i][4])
-       url=URLDecodeFormArgs(arglist[i]+4);
+    *colon=0;
 
-    if(!strncmp("proto=",arglist[i],6) && arglist[i][6])
-       proto=URLDecodeFormArgs(arglist[i]+6);
-    if(!strncmp("host=",arglist[i],5) && arglist[i][5])
-       host=URLDecodeFormArgs(arglist[i]+5);
-    if(!strncmp("port=",arglist[i],5) && arglist[i][5])
-       port=URLDecodeFormArgs(arglist[i]+5);
-    if(!strncmp("path=",arglist[i],5) && arglist[i][5])
-       path=URLDecodeFormArgs(arglist[i]+5);
-    if(!strncmp("args=",arglist[i],5) && arglist[i][5])
-       args=URLDecodeFormArgs(arglist[i]+5);
-
-    if(!strncmp("proto_other=",arglist[i],12) && arglist[i][12])
-       proto_other=URLDecodeFormArgs(arglist[i]+12);
-    if(!strncmp("host_other=",arglist[i],11) && arglist[i][11])
-       host_other=URLDecodeFormArgs(arglist[i]+11);
-    if(!strncmp("port_other=",arglist[i],11) && arglist[i][11])
-       port_other=URLDecodeFormArgs(arglist[i]+11);
-    if(!strncmp("path_other=",arglist[i],11) && arglist[i][11])
-       path_other=URLDecodeFormArgs(arglist[i]+11);
-    if(!strncmp("args_other=",arglist[i],11) && arglist[i][11])
-       args_other=URLDecodeFormArgs(arglist[i]+11);
+    colon=strchr(slash+2,':');
+    slash=strchr(slash+2,'/');
    }
 
- free(arglist[0]);
- free(arglist);
- free(form);
-
- /* Sort out a URL from the mess of arguments. */
-
- if(url)
+ if(colon && slash && colon<slash)
    {
-    char *colon;
-    URL *Url=SplitURL(url);
-
-    if(proto_other) free(proto_other);
-    if(host_other) free(host_other);
-    if(port_other) free(port_other);
-    if(path_other) free(path_other);
-    if(args_other) free(args_other);
-
-    proto=(char*)realloc((void*)proto,strlen(Url->proto)+1);
-    strcpy(proto,Url->proto);
-
-    colon=strchr(Url->host,':');
-
-    host=(char*)realloc((void*)host,strlen(Url->host)+1);
-    strcpy(host,Url->host);
-    if(colon)
-       host[colon-Url->host]=0;
-
-    if(colon)
-      {
-       port=(char*)realloc((void*)port,strlen(colon+1)+1);
-       strcpy(port,colon+1);
-      }
-
-    path=(char*)realloc((void*)path,strlen(Url->path)+1);
-    strcpy(path,Url->path);
-
-    if(Url->args)
-      {
-       args=(char*)realloc((void*)args,strlen(Url->args)+1);
-       strcpy(args,Url->args);
-      }
-
-    FreeURL(Url);
-   }
- else
-   {
-    int urllen;
-
-    if(proto && !strcmp(proto,"OTHER") && proto_other)
-      {free(proto); proto=proto_other;}
+    *colon=0;
+    if(slash==colon+1)
+       port=":";
     else
-      {if(proto_other) free(proto_other);}
-
-    if(host && !strcmp(host,"OTHER") && host_other)
-      {free(host); host=host_other;}
-    else
-      {if(host_other) free(host_other);}
-
-    if(port && !strcmp(port,"OTHER") && port_other)
-      {free(port); port=(char*)malloc(strlen(port_other)+1); strcpy(port,":"); strcat(port,port_other);}
-    else
-      {if(port_other) free(port_other);}
-
-    if(path && !strcmp(path,"OTHER") && path_other)
-      {free(path); path=path_other;}
-    else
-      {if(path_other) free(path_other);}
-
-    if(args && !strcmp(args,"OTHER") && args_other)
-      {free(args); args=(char*)malloc(strlen(args_other)+1); strcpy(args,"?"); strcat(args,args_other);}
-    else
-      {if(args_other) free(args_other);}
-
-    if(!proto)
-       proto=strcpy(malloc(2),"*");
-
-    if(!host)
-       host=strcpy(malloc(2),"*");
-
-    if(!path)
-       path=strcpy(malloc(2),"*");
-
-    urllen=4;
-    urllen+=strlen(proto);
-    urllen+=strlen(host);
-    if(port)
-       urllen+=strlen(port);
-    urllen+=strlen(path);
-    if(args)
-       urllen+=strlen(args);
-
-    url=(char*)realloc((void*)url,urllen);
-
-    strcpy(url,proto);
-    strcat(url,"://");
-    strcat(url,host);
-    if(port)
-       strcat(url,port);
-    strcat(url,path);
-    if(args)
-       strcat(url,args);
+       port=colon+1;
    }
 
- for(i=0;url[i];i++)
-    if(url[i]=='*')
-       wildcard++;
+ *slash=0;
 
- if(!wildcard)
-    Url=SplitURL(url);
+ path=slash+1;
+
+ ques=strchr(path,'?');
+
+ if(ques)
+   {
+    if(*(ques+1)==0)
+       args="?";
+    else
+       args=ques+1;
+
+    *ques=0;
+   }
+
+ if(strchr(url,'*') || (port && *port==':') || (args && *args=='?'))
+    wildcard=1;
 
  /* Display the HTML */
 
- file=OpenLanguageFile("README.CONF.txt");
+ file=OpenLanguageFile("messages/README.CONF.txt");
 
  if(file==-1)
    {
-    HTMLMessage(fd,200,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
+    HTMLMessage(fd,500,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
                 "error","Cannot open README.CONF.txt",
                 NULL);
+
+    free(copy);
+
     return;
    }
 
+ init_io(file);
+
  HTMLMessageHead(fd,200,"WWWOFFLE Configuration URL Page",
                  NULL);
 
@@ -876,141 +830,125 @@ static void ConfigurationURLPage(int fd,char *request_args)
                  "url",url,
                  "proto",proto,
                  "host",host,
-                 "port",port&&*port==':'?port+1:port,
+                 "port",port,
                  "path",path,
-                 "args",args&&*args=='?'?args+1:args,
+                 "args",args,
                  NULL);
 
- /* Loop through all of the specified ConfigItems. */
+ free(copy);
 
- for(configitem=0;configitem<sizeof(URLConfigItems)/sizeof(URLConfigItems[0]);configitem++)
+ /* Loop through all of the ConfigItems and find those that take a URL argument. */
+
+ for(s=0;s<CurrentConfig.nsections;s++)
    {
-    char *template=NULL,*description=NULL,*current=NULL;
-
-    for(s=0,i=0;s<CurrentConfig.nsections;s++)
-      {
-       for(i=0;i<CurrentConfig.sections[s]->nitemdefs;i++)
-          if(URLConfigItems[configitem]==CurrentConfig.sections[s]->itemdefs[i].item)
-             break;
-
-       if(i!=CurrentConfig.sections[s]->nitemdefs)
-          break;
-      }
-
-    if(s==CurrentConfig.nsections)
-       continue;
-
-    if(s!=last_s)
-      {
-       lseek(file,0,SEEK_SET);
-
-       while((line1=read_line(file,line1)))
-         {
-          line1[strlen(line1)-1]=0;
-
-          if(!strncmp(line1,"SECTION",7) && !strcmp(line1+8,CurrentConfig.sections[s]->name))
-            {
-             line1=read_line(file,line1);
-             if(line1)
-                line1[strlen(line1)-1]=0;
-             description=line1;
-             seekpos=lseek(file,0,SEEK_CUR);
-             if(seekpos>READ_BUFFER_SIZE)
-                seekpos-=READ_BUFFER_SIZE;
-             else
-                seekpos=0;
-             break;
-            }
-         }
-
-       last_s=s;
-      }
-    else
-      {
-       lseek(file,seekpos,SEEK_SET);
-       init_buffer(file);
-
-       while((line1=read_line(file,line1)))
-         {
-          line1[strlen(line1)-1]=0;
-
-          if(!strncmp(line1,"SECTION",7) && !strcmp(line1+8,CurrentConfig.sections[s]->name))
-             break;
-         }
-      }
+    lseek(file,0,SEEK_SET);  /* go back to the start of the file */
+    reinit_io(file);
 
     while((line1=read_line(file,line1)))
       {
        line1[strlen(line1)-1]=0;
 
-       if(!strncmp(line1,"ITEM",4) && !strcmp(line1+5,CurrentConfig.sections[s]->itemdefs[i].name))
+       if(!strncmp(line1,"SECTION",7) && !strcmp(line1+8,CurrentConfig.sections[s]->name))
          {
-          line1=read_line(file,line1);
-          if(line1)
-             line1[strlen(line1)-1]=0;
-          line2=read_line(file,line2);
-          if(line2)
-             line2[strlen(line2)-1]=0;
-          template=line1;
-          description=line2;
+          seekpos=lseek(file,0,SEEK_CUR);
+          if(seekpos>READ_BUFFER_SIZE)
+             seekpos-=READ_BUFFER_SIZE;
+          else
+             seekpos=0;
           break;
          }
       }
 
-    if(!wildcard)
-      {
-       if(*URLConfigItems[configitem] &&
-          (*URLConfigItems[configitem])->itemdef->key_type!=String)
+    for(i=0;i<CurrentConfig.sections[s]->nitemdefs;i++)
+       if(CurrentConfig.sections[s]->itemdefs[i].url_type ||
+          CurrentConfig.sections[s]->itemdefs[i].key_type==UrlSpecification)
          {
-          int i;
+          char *template=NULL,*description=NULL,*current=NULL;
 
-          for(i=0;i<(*URLConfigItems[configitem])->nentries;i++)
-             if((*URLConfigItems[configitem])->itemdef->url_type)
+          lseek(file,seekpos,SEEK_SET);  /* go back only to the start of the current section */
+          reinit_io(file);
+
+          while((line1=read_line(file,line1)))
+            {
+             line1[strlen(line1)-1]=0;
+
+             if(!strncmp(line1,"SECTION",7) && !strcmp(line1+8,CurrentConfig.sections[s]->name))
+                break;
+            }
+
+          while((line1=read_line(file,line1)))
+            {
+             line1[strlen(line1)-1]=0;
+
+             if(!strncmp(line1,"SECTION",7))
+                break;
+             if(!strncmp(line1,"ITEM",4) && !strcmp(line1+5,CurrentConfig.sections[s]->itemdefs[i].name))
                {
-                if(!(*URLConfigItems[configitem])->url[i])
-                  break;
-                else if(MatchUrlSpecification((*URLConfigItems[configitem])->url[i],Url->proto,Url->host,Url->path,Url->args))
-                   break;
+                line1=read_line(file,line1);
+                if(line1)
+                   line1[strlen(line1)-1]=0;
+                line2=read_line(file,line2);
+                if(line2)
+                   line2[strlen(line2)-1]=0;
+                template=line1;
+                description=line2;
+                break;
                }
-             else
+            }
+
+          if(!wildcard)
+            {
+             URL *Url=SplitURL(url);
+
+             if(*CurrentConfig.sections[s]->itemdefs[i].item &&
+                CurrentConfig.sections[s]->itemdefs[i].key_type!=String)
                {
-                if(MatchUrlSpecification((*URLConfigItems[configitem])->key[i].urlspec,Url->proto,Url->host,Url->path,Url->args))
-                   break;
+                int j;
+
+                for(j=0;j<(*CurrentConfig.sections[s]->itemdefs[i].item)->nentries;j++)
+                   if(CurrentConfig.sections[s]->itemdefs[i].url_type)
+                     {
+                      if(!(*CurrentConfig.sections[s]->itemdefs[i].item)->url[j])
+                         break;
+                      else if(MatchUrlSpecification((*CurrentConfig.sections[s]->itemdefs[i].item)->url[j],Url->proto,Url->host,Url->path,Url->args))
+                         break;
+                     }
+                   else
+                     {
+                      if(MatchUrlSpecification((*CurrentConfig.sections[s]->itemdefs[i].item)->key[j].urlspec,Url->proto,Url->host,Url->path,Url->args))
+                         break;
+                     }
+
+                if(j!=(*CurrentConfig.sections[s]->itemdefs[i].item)->nentries)
+                   current=ConfigEntryString(*CurrentConfig.sections[s]->itemdefs[i].item,j);
                }
 
-          if(i!=(*URLConfigItems[configitem])->nentries)
-             current=ConfigEntryString(*URLConfigItems[configitem],i);
+             FreeURL(Url);
+            }
+
+          HTMLMessageBody(fd,"ConfigurationUrl-Body",
+                          "section",CurrentConfig.sections[s]->name,
+                          "item",CurrentConfig.sections[s]->itemdefs[i].name,
+                          "template",template,
+                          "description",description,
+                          "current",current,
+                          NULL);
+
+          if(current)
+             free(current);
          }
-      }
-
-    HTMLMessageBody(fd,"ConfigurationUrl-Body",
-                    "section",CurrentConfig.sections[s]->name,
-                    "item",CurrentConfig.sections[s]->itemdefs[i].name,
-                    "template",template,
-                    "description",description,
-                    "current",current,
-                    NULL);
-
-    if(current)
-       free(current);
    }
 
  HTMLMessageBody(fd,"ConfigurationUrl-Tail",
                  NULL);
 
- if(proto) free(proto);
- if(host) free(host);
- if(port) free(port);
- if(path) free(path);
- if(args) free(args);
-
- if(Url)
-    FreeURL(Url);
-
  if(line1)
     free(line1);
  if(line2)
     free(line2);
+
+ finish_io(file);
+ close(file);
 }
 
 
