@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/src/RCS/configrdwr.c 1.39 2002/03/09 11:47:14 amb Exp $
+  $Header: /home/amb/wwwoffle/src/RCS/configrdwr.c 1.58 2003/11/03 16:34:50 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.7a.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.8a.
   Configuration file reading and writing functions.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1997,98,99,2000,01,02 Andrew M. Bishop
+  This file Copyright 1997,98,99,2000,01,02,03 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -20,31 +20,50 @@
 #include <string.h>
 #include <ctype.h>
 
+#if TIME_WITH_SYS_TIME
+# include <sys/time.h>
+# include <time.h>
+#else
+# if HAVE_SYS_TIME_H
+#  include <sys/time.h>
+# else
+#  include <time.h>
+# endif
+#endif
+
 #include <limits.h>
-#include <unistd.h>
 #include <sys/types.h>
+#include <unistd.h>
 #include <sys/stat.h>
+#include <fcntl.h>
 #include <pwd.h>
 #include <grp.h>
 
-#include "configpriv.h"
-#include "wwwoffle.h"
-#include "errors.h"
+#include "io.h"
 #include "misc.h"
+#include "errors.h"
+#include "configpriv.h"
+#include "config.h"
 
 
+/*+ Need this for Win32 to use binary mode +*/
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
+
 #ifndef PATH_MAX
 #define PATH_MAX 4096
 #endif
 
+
 /* Local functions */
 
 static char *filename_or_symlink_target(char *name);
 
-static char *InitParser(void);
-static char *ParseLine(char **line);
-static char *ParseItem(char *line,char **url_str,char **key_str,char **val_str);
-static char *ParseEntry(ConfigItemDef *itemdef,ConfigItem *item,char *url_str,char *key_str,char *val_str);
+static /*@null@*/ char *InitParser(void);
+static /*@null@*/ char *ParseLine(/*@out@*/ /*@null@*/ char **line);
+static /*@null@*/ char *ParseItem(char *line,/*@out@*/ char **url_str,/*@out@*/ char **key_str,/*@out@*/ char **val_str);
+static /*@null@*/ char *ParseEntry(ConfigItemDef *itemdef,/*@out@*/ ConfigItem *item,/*@null@*/ char *url_str,char *key_str,/*@null@*/ char *val_str);
 
 static int isanumber(const char *string);
 
@@ -69,11 +88,11 @@ typedef enum _ParserState
 ParserState;
 
 static char *parse_name;        /*+ The name of the configuration file. +*/
-static FILE *parse_file;        /*+ The file handle of the configuration file. +*/
+static int parse_file;          /*+ The file descriptor of the configuration file. +*/
 static int parse_line;          /*+ The line number in the configuration file. +*/
 
 static char *parse_name_org;    /*+ The name of the original configuration file when parsing an included one. +*/
-static FILE *parse_file_org;    /*+ The file handle of the original configuration file when parsing an included one. +*/
+static int parse_file_org;      /*+ The file descriptor of the original configuration file when parsing an included one. +*/
 static int parse_line_org;      /*+ The line number in the original configuration file when parsing an included one. +*/
 
 static int parse_section;       /*+ The current section of the configuration file. +*/
@@ -97,10 +116,11 @@ char *ReadConfigFile(void)
  errmsg=InitParser();
 
  if(!errmsg)
+   {
+    char *line=NULL;
+
     do
       {
-       char *line=NULL;
-
        if((errmsg=ParseLine(&line)))
           break;
 
@@ -122,6 +142,21 @@ char *ReadConfigFile(void)
       }
     while(parse_state!=Finished);
 
+    if(line)
+       free(line);
+   }
+
+ if(parse_file!=-1)
+   {
+    finish_io(parse_file);
+    close(parse_file);
+   }
+ if(parse_file_org!=-1)
+   {
+    finish_io(parse_file_org);
+    close(parse_file_org);
+   }
+
  if(errmsg)
     RestoreBackupConfigFile();
  else
@@ -167,10 +202,10 @@ char *ModifyConfigFile(int section,int item,char *newentry,char *preventry,char
 {
  char *errmsg=NULL;
  char **names=(char**)calloc((1+CurrentConfig.nsections),sizeof(char*));
- FILE *file=NULL,*file_orig=NULL;
+ int file=-1,file_org=-1;
  ConfigItem dummy=NULL;
  int matched=0;
- int s;
+ int s,rename_failed=0;
 
  /* Initialise the parser and open the new file. */
 
@@ -182,21 +217,25 @@ char *ModifyConfigFile(int section,int item,char *newentry,char *preventry,char
 
     strcat(names[0],".new");
 
-    file=fopen(names[0],"w");
-    if(!file)
+    file=open(names[0],O_WRONLY|O_CREAT|O_TRUNC|O_BINARY);
+
+    if(file==-1)
       {
        errmsg=(char*)malloc(64+strlen(names[0]));
        sprintf(errmsg,"Cannot open the configuration file '%s' for writing.",names[0]);
       }
+    else
+       init_io(file);
    }
 
  /* Parse the file */
 
  if(!errmsg)
+   {
+    char *line=NULL;
+
     do
       {
-       char *line=NULL;
-
        if((errmsg=ParseLine(&line)))
           break;
 
@@ -207,7 +246,7 @@ char *ModifyConfigFile(int section,int item,char *newentry,char *preventry,char
 
           /* Insert a new entry for a non-existing item. */
 
-          if(!preventry && !sameentry && !nextentry && !matched &&
+          if(newentry && !preventry && !sameentry && !nextentry && !matched &&
              (parse_state==InsideSectionCurly || parse_state==InsideSectionIncluded))
             {
              char *copyentry=(char*)malloc(strlen(newentry)+1);
@@ -221,9 +260,11 @@ char *ModifyConfigFile(int section,int item,char *newentry,char *preventry,char
                                    url_str,key_str,val_str)))
                 break;
 
-             fprintf(file,"\n# WWWOFFLE Configuration Edit Inserted: %s\n %s\n\n",RFC822Date(time(NULL),0),newentry);
+             write_formatted(file,"\n# WWWOFFLE Configuration Edit Inserted: %s\n %s\n\n",RFC822Date(time(NULL),0),newentry);
 
              matched=1;
+
+             free(copyentry);
             }
 
           copy=(char*)malloc(strlen(line)+1);
@@ -258,10 +299,12 @@ char *ModifyConfigFile(int section,int item,char *newentry,char *preventry,char
                                       url_str,key_str,val_str)))
                    break;
 
-                fprintf(file,"\n# WWWOFFLE Configuration Edit Inserted: %s\n %s\n\n",RFC822Date(time(NULL),0),newentry);
-                fprintf(file,"%s",line);
+                write_formatted(file,"\n# WWWOFFLE Configuration Edit Inserted: %s\n %s\n\n",RFC822Date(time(NULL),0),newentry);
+                write_string(file,line);
 
                 matched=1;
+
+                free(copyentry);
                }
 
              /* Insert a new entry after the current one */
@@ -279,17 +322,19 @@ char *ModifyConfigFile(int section,int item,char *newentry,char *preventry,char
                                       url_str,key_str,val_str)))
                    break;
 
-                fprintf(file,"%s",line);
-                fprintf(file,"\n# WWWOFFLE Configuration Edit Inserted: %s\n %s\n\n",RFC822Date(time(NULL),0),newentry);
+                write_string(file,line);
+                write_formatted(file,"\n# WWWOFFLE Configuration Edit Inserted: %s\n %s\n\n",RFC822Date(time(NULL),0),newentry);
 
                 matched=1;
+
+                free(copyentry);
                }
 
              /* Delete an entry */
 
              else if(!newentry && sameentry && !strcmp(thisentry,sameentry))
                {
-                fprintf(file,"\n# WWWOFFLE Configuration Edit Deleted: %s\n#%s",RFC822Date(time(NULL),0),line);
+                write_formatted(file,"\n# WWWOFFLE Configuration Edit Deleted: %s\n#%s",RFC822Date(time(NULL),0),line);
 
                 matched=1;
                }
@@ -316,47 +361,57 @@ char *ModifyConfigFile(int section,int item,char *newentry,char *preventry,char
                                       url_str,key_str,val_str)))
                    break;
 
-                fprintf(file,"# WWWOFFLE Configuration Edit Changed: %s\n#%s",RFC822Date(time(NULL),0),line);
-                fprintf(file," %s\n",newentry);
+                write_formatted(file,"# WWWOFFLE Configuration Edit Changed: %s\n#%s",RFC822Date(time(NULL),0),line);
+                write_formatted(file," %s\n",newentry);
 
                 matched=1;
+
+                free(copyentry);
                }
              else
-                fprintf(file,"%s",line);
+                write_string(file,line);
 
              free(thisentry);
             }
           else
-             fprintf(file,"%s",line);
+             write_string(file,line);
 
           free(copy);
          }
        else if(line)
-          fprintf(file,"%s",line);
+          write_string(file,line);
 
-       if(parse_state==StartSectionIncluded && !file_orig)
+       if(parse_state==StartSectionIncluded && file_org==-1)
          {
           names[parse_section+1]=filename_or_symlink_target(parse_name);
           strcat(names[parse_section+1],".new");
 
-          file_orig=file;
-          file=fopen(names[parse_section+1],"w");
-          if(!file)
+          file_org=file;
+          file=open(names[parse_section+1],O_WRONLY|O_CREAT|O_TRUNC|O_BINARY);
+
+          if(file==-1)
             {
-             char *errmsg=(char*)malloc(48+strlen(names[0]));
+             errmsg=(char*)malloc(48+strlen(names[parse_section+1]));
              sprintf(errmsg,"Cannot open the included file '%s' for writing.",names[parse_section+1]);
              break;
             }
+
+          init_io(file);
          }
-       else if(parse_state==StartSectionSquare && file_orig)
+       else if(parse_state==StartSectionSquare && file_org!=-1)
          {
-          fclose(file);
-          file=file_orig;
-          file_orig=NULL;
+          finish_io(file);
+          close(file);
+          file=file_org;
+          file_org=-1;
          }
       }
     while(parse_state!=Finished);
 
+    if(line)
+       free(line);
+   }
+
  if(!errmsg && !matched && (preventry || sameentry || nextentry))
    {
     char *whichentry=sameentry?sameentry:preventry?preventry:nextentry;
@@ -365,16 +420,33 @@ char *ModifyConfigFile(int section,int item,char *newentry,char *preventry,char
     sprintf(errmsg,"No entry to match '%s' was found to make the change.",whichentry);
    }
 
- if(file)
-    fclose(file);
- if(file_orig)
-    fclose(file_orig);
+ if(file!=-1)
+   {
+    finish_io(file);
+    close(file);
+   }
+ if(file_org!=-1)
+   {
+    finish_io(file_org);
+    close(file_org);
+   }
+ if(parse_file!=-1)
+   {
+    finish_io(parse_file);
+    close(parse_file);
+   }
+ if(parse_file_org!=-1)
+   {
+    finish_io(parse_file_org);
+    close(parse_file_org);
+   }
 
  for(s=0;s<=CurrentConfig.nsections;s++)
     if(names[s])
       {
        if(!errmsg)
          {
+          struct stat buf;
           char *name=(char*)malloc(strlen(names[s])+1);
           char *name_bak=(char*)malloc(strlen(names[s])+16);
 
@@ -384,17 +456,38 @@ char *ModifyConfigFile(int section,int item,char *newentry,char *preventry,char
           strcpy(name_bak,names[s]);
           strcpy(name_bak+strlen(name_bak)-3,"bak");
 
-          rename(name,name_bak);
-          rename(names[s],name);
+          if(!stat(name,&buf))
+            {
+             chown(names[s],buf.st_uid,buf.st_gid);
+             chmod(names[s],buf.st_mode&(~S_IFMT));
+            }
+
+          if(rename(name,name_bak))
+            {
+             rename_failed++;
+             PrintMessage(Warning,"Cannot rename '%s' to '%s' when modifying configuration entry [%!s].",name,name_bak);
+            }
+          if(rename(names[s],name))
+            {
+             rename_failed++;
+             PrintMessage(Warning,"Cannot rename '%s' to '%s' when modifying configuration entry [%!s].",name[s],name);
+            }
 
-          free(names[s]);
           free(name);
           free(name_bak);
          }
        else
           unlink(names[s]);
+
+       free(names[s]);
       }
 
+ if(rename_failed)
+   {
+    errmsg=(char*)malloc(120);
+    sprintf(errmsg,"There were problems renaming files, check the error log (this might stop the change you tried to make).");
+   }
+
  FreeConfigItem(dummy);
 
  free(names);
@@ -411,7 +504,7 @@ char *ModifyConfigFile(int section,int item,char *newentry,char *preventry,char
   char *name The file name that may be a symbolic link.
   ++++++++++++++++++++++++++++++++++++++*/
 
-char *filename_or_symlink_target(char *name)
+static char *filename_or_symlink_target(char *name)
 {
  struct stat buf;
  char linkname[PATH_MAX+1];
@@ -463,23 +556,25 @@ char *filename_or_symlink_target(char *name)
 static char *InitParser(void)
 {
  parse_name=CurrentConfig.name;
- parse_file=fopen(parse_name,"r");
+ parse_file=open(parse_name,O_RDONLY|O_BINARY);
  parse_line=0;
 
- if(!parse_file)
+ parse_name_org=NULL;
+ parse_file_org=-1;
+ parse_line_org=0;
+
+ parse_section=-1;
+ parse_item=-1;
+ parse_state=OutsideSection;
+
+ if(parse_file==-1)
    {
     char *errmsg=(char*)malloc(64+strlen(parse_name));
     sprintf(errmsg,"Cannot open the configuration file '%s' for reading.",parse_name);
     return(errmsg);
    }
 
- parse_name_org=NULL;
- parse_file_org=NULL;
- parse_line_org=0;
-
- parse_section=-1;
- parse_item=-1;
- parse_state=OutsideSection;
+ init_io(parse_file);
 
  return(NULL);
 }
@@ -499,7 +594,7 @@ static char *ParseLine(char **line)
 
  /* Read from the line and make a copy */
 
- *line=fgets_realloc(*line,parse_file);
+ *line=read_line(parse_file,*line);
 
  parse_line++;
 
@@ -513,11 +608,12 @@ static char *ParseLine(char **line)
        parse_state=Finished;
     else if(parse_state==StartSectionIncluded || parse_state==InsideSectionIncluded)
       {
-       fclose(parse_file);
+       finish_io(parse_file);
+       close(parse_file);
        free(parse_name);
 
        parse_file=parse_file_org;
-       parse_file_org=NULL;
+       parse_file_org=-1;
 
        parse_name=parse_name_org;
        parse_name_org=NULL;
@@ -629,35 +725,49 @@ static char *ParseLine(char **line)
           errmsg=(char*)malloc(48);
           sprintf(errmsg,"End of section '%c' has trailing junk.",*l);
          }
+       else if(*l!='#' && *l!=0 && strchr(l,'/'))
+         {
+          errmsg=(char*)malloc(64);
+          sprintf(errmsg,"Included file must be in same directory (no '/').");
+         }
        else if(*l!='#' && *l!=0)
          {
           char *rr;
+          char *inc_parse_name;
+          int inc_parse_file;
 
-          parse_name_org=parse_name;
-          parse_file_org=parse_file;
-          parse_line_org=parse_line;
+          inc_parse_name=(char*)malloc(strlen(CurrentConfig.name)+(r-l)+1);
 
-          parse_name=(char*)malloc(strlen(CurrentConfig.name)+(r-l)+1);
+          strcpy(inc_parse_name,CurrentConfig.name);
 
-          strcpy(parse_name,CurrentConfig.name);
-
-          rr=parse_name+strlen(parse_name)-1;
-          while(rr>parse_name && *rr!='/')
+          rr=inc_parse_name+strlen(inc_parse_name)-1;
+          while(rr>inc_parse_name && *rr!='/')
              rr--;
 
           strncpy(rr+1,l,r-l);
           *((rr+1)+(r-l))=0;
 
-          parse_line=0;
-          parse_file=fopen(parse_name,"r");
+          inc_parse_file=open(inc_parse_name,O_RDONLY|O_BINARY);
 
-          if(!parse_file)
+          if(inc_parse_file==-1)
             {
-             errmsg=(char*)malloc(48+strlen(parse_name));
-             sprintf(errmsg,"Cannot open included file '%s' for reading.",parse_name);
+             errmsg=(char*)malloc(48+strlen(inc_parse_name));
+             sprintf(errmsg,"Cannot open the included file '%s' for reading.",inc_parse_name);
             }
           else
+            {
+             init_io(inc_parse_file);
+
              parse_state=StartSectionIncluded;
+
+             parse_name_org=parse_name;
+             parse_file_org=parse_file;
+             parse_line_org=parse_line;
+
+             parse_name=inc_parse_name;
+             parse_file=inc_parse_file;
+             parse_line=0;
+            }
          }
       }
 
@@ -667,14 +777,6 @@ static char *ParseLine(char **line)
       }
    }
 
- /* Done, tidy up and return. */
-
- if(*line && (errmsg || parse_state==Finished))
-   {
-    free(*line);
-    *line=NULL;
-   }
-
  return(errmsg);
 }
 
@@ -695,6 +797,10 @@ static char *ParseLine(char **line)
 
 static char *ParseItem(char *line,char **url_str,char **key_str,char **val_str)
 {
+ *url_str=NULL;
+ *key_str=NULL;
+ *val_str=NULL;
+
  if(parse_state==InsideSectionCurly || parse_state==InsideSectionIncluded)
    {
     char *url=NULL,*key=NULL,*val=NULL;
@@ -754,6 +860,27 @@ static char *ParseItem(char *line,char **url_str,char **key_str,char **val_str)
              if(*ll && *ll!='=' && !isspace(*ll))
                 continue;
             }
+          else if(CurrentConfig.sections[parse_section]->itemdefs[parse_item].key_type==UrlSpecification)
+            {
+             char *equal;
+
+             ll=key;
+
+             while((equal=strchr(ll,'=')))
+               {
+                ll=equal;
+                if(--equal>key && isspace(*equal))
+                  {
+                   while(isspace(*equal))
+                      *equal--=0;
+                   break;
+                  }
+                ll++;
+               }
+
+             while(*ll && *ll!='=' && !isspace(*ll))
+                ll++;
+            }
           else
             {
              ll=key;
@@ -840,9 +967,12 @@ static char *ParseItem(char *line,char **url_str,char **key_str,char **val_str)
 
 static char *ParseEntry(ConfigItemDef *itemdef,ConfigItem *item,char *url_str,char *key_str,char *val_str)
 {
- UrlSpec *url;
+ UrlSpec *url=NULL;
  KeyOrValue key,val;
- char *errmsg;
+ char *errmsg=NULL;
+
+ key.string=NULL;
+ val.string=NULL;
 
  if(itemdef->same_key==0 && itemdef->url_type==0 && (*item) && (*item)->nentries)
    {
@@ -863,13 +993,17 @@ static char *ParseEntry(ConfigItemDef *itemdef,ConfigItem *item,char *url_str,ch
       {
        errmsg=(char*)malloc(32+strlen(key_str));
        sprintf(errmsg,"Unexpected key string: '%s'.",key_str);
+       if(url) free(url);
        return(errmsg);
       }
     key.string=itemdef->name;
    }
  else
     if((errmsg=ParseKeyOrValue(key_str,itemdef->key_type,&key)))
+      {
+       if(url) free(url);
        return(errmsg);
+      }
 
  if(!val_str)
     val.string=NULL;
@@ -877,10 +1011,19 @@ static char *ParseEntry(ConfigItemDef *itemdef,ConfigItem *item,char *url_str,ch
     val.string=NULL;
  else
     if((errmsg=ParseKeyOrValue(val_str,itemdef->val_type,&val)))
+      {
+       if(url) free(url);
+       FreeKeyOrValue(&key,itemdef->key_type);
        return(errmsg);
+      }
 
  if(!item)
+   {
+    if(url) free(url);
+    FreeKeyOrValue(&key,itemdef->key_type);
+    FreeKeyOrValue(&val,itemdef->val_type);
     return(NULL);
+   }
 
  if(!(*item))
    {
@@ -1074,7 +1217,7 @@ char *ParseKeyOrValue(char *text,ConfigType type,KeyOrValue *pointer)
     else
       {
        pointer->integer=atoi(text);
-       if(pointer->integer<0)
+       if(pointer->integer<-1)
          {errmsg=(char*)malloc(48);sprintf(errmsg,"Invalid cache size %d.",pointer->integer);}
       }
     break;
@@ -1185,7 +1328,7 @@ char *ParseKeyOrValue(char *text,ConfigType type,KeyOrValue *pointer)
     else if(!isanumber(text) || *text!='0')
       {errmsg=(char*)malloc(48+strlen(text));sprintf(errmsg,"Expecting a file permissions mode, got '%s'.",text);}
     else
-       pointer->integer=strtol(text,NULL,8);
+       sscanf(text,"%o",&pointer->integer);
     break;
 
    case MIMEType:
@@ -1208,7 +1351,7 @@ char *ParseKeyOrValue(char *text,ConfigType type,KeyOrValue *pointer)
        break;
       }
 
-    /* fall through */
+    /*@fallthrough@*/
 
    case Host:
     if(!*text)
@@ -1269,7 +1412,7 @@ char *ParseKeyOrValue(char *text,ConfigType type,KeyOrValue *pointer)
        break;
       }
 
-    /* fall through */
+    /*@fallthrough@*/
 
    case HostAndPort:
     if(!*text)
@@ -1325,8 +1468,8 @@ char *ParseKeyOrValue(char *text,ConfigType type,KeyOrValue *pointer)
          {
           pointer->string=(char*)malloc(strlen(text)+1);
           strcpy(pointer->string,text);
-          FreeURL(tempUrl);
          }
+       FreeURL(tempUrl);
       }
     break;
 
@@ -1338,47 +1481,49 @@ char *ParseKeyOrValue(char *text,ConfigType type,KeyOrValue *pointer)
        char *p,*orgtext=text;
 
        pointer->urlspec=(UrlSpec*)malloc(sizeof(UrlSpec));
+       pointer->urlspec->null=0;
        pointer->urlspec->negated=0;
+       pointer->urlspec->nocase=0;
        pointer->urlspec->proto=0;
        pointer->urlspec->host=0;
        pointer->urlspec->port=-1;
        pointer->urlspec->path=0;
        pointer->urlspec->args=0;
 
-       /* ! */
+       /* !~ */
 
-       if(*text=='!')
+       while(*text)
          {
-          pointer->urlspec->negated=1;
+          if(*text=='!')
+             pointer->urlspec->negated=1;
+          else if(*text=='~')
+             pointer->urlspec->nocase=1;
+          else
+             break;
+
           text++;
          }
 
        /* protocol */
 
        if(!strncmp(text,"*://",4))
-         {
           p=text+4;
-          pointer->urlspec->proto=0;
-         }
        else if(!strncmp(text,"://",3))
-         {
           p=text+3;
-          pointer->urlspec->proto=0;
-         }
        else if((p=strstr(text,"://")))
          {
-          pointer->urlspec=(UrlSpec*)realloc((void*)pointer->urlspec,
-                                             sizeof(UrlSpec)+
-                                             (p-text+1));
-
           pointer->urlspec->proto=sizeof(UrlSpec);
 
+          pointer->urlspec=(UrlSpec*)realloc((void*)pointer->urlspec,
+                                             pointer->urlspec->proto+(p-text)+1);
+
           strncpy(UrlSpecProto(pointer->urlspec),text,p-text);
           *(UrlSpecProto(pointer->urlspec)+(p-text))=0;
           p+=3;
          }
        else
-         {errmsg=(char*)malloc(64+strlen(orgtext));sprintf(errmsg,"Expecting a URL-SPECIFICATION, got this '%s'.",orgtext); break;}
+         {errmsg=(char*)malloc(64+strlen(orgtext));sprintf(errmsg,"Expecting a URL-SPECIFICATION, got this '%s'.",orgtext);
+          free(pointer->urlspec); break;}
 
        if(pointer->urlspec->proto)
          {
@@ -1421,13 +1566,11 @@ char *ParseKeyOrValue(char *text,ConfigType type,KeyOrValue *pointer)
          {
           char *q;
 
-          pointer->urlspec=(UrlSpec*)realloc((void*)pointer->urlspec,
-                                             sizeof(UrlSpec)+
-                                             (pointer->urlspec->proto?1+strlen(UrlSpecProto(pointer->urlspec)):0)+
-                                             (p-text+1));
+          pointer->urlspec->host=(unsigned short)(sizeof(UrlSpec)+
+                                                  (pointer->urlspec->proto?1+strlen(UrlSpecProto(pointer->urlspec)):0));
 
-          pointer->urlspec->host=sizeof(UrlSpec)+
-                                 (pointer->urlspec->proto?1+strlen(UrlSpecProto(pointer->urlspec)):0);
+          pointer->urlspec=(UrlSpec*)realloc((void*)pointer->urlspec,
+                                             pointer->urlspec->host+(p-text)+1);
 
           strncpy(UrlSpecHost(pointer->urlspec),text,p-text);
           *(UrlSpecHost(pointer->urlspec)+(p-text))=0;
@@ -1457,7 +1600,8 @@ char *ParseKeyOrValue(char *text,ConfigType type,KeyOrValue *pointer)
           p=text+2;
          }
        else if(*text==':')
-         {errmsg=(char*)malloc(64+strlen(orgtext));sprintf(errmsg,"Expecting a URL-SPECIFICATION, got this '%s'.",orgtext); break;}
+         {errmsg=(char*)malloc(64+strlen(orgtext));sprintf(errmsg,"Expecting a URL-SPECIFICATION, got this '%s'.",orgtext);
+          free(pointer->urlspec); break;}
 
        text=p;
 
@@ -1479,7 +1623,8 @@ char *ParseKeyOrValue(char *text,ConfigType type,KeyOrValue *pointer)
              pointer->urlspec->path=1;
          }
        else
-         {errmsg=(char*)malloc(64+strlen(orgtext));sprintf(errmsg,"Expecting a URL-SPECIFICATION, got this '%s'.",orgtext); break;}
+         {errmsg=(char*)malloc(64+strlen(orgtext));sprintf(errmsg,"Expecting a URL-SPECIFICATION, got this '%s'.",orgtext);
+          free(pointer->urlspec); break;}
 
        if(pointer->urlspec->path)
          {
@@ -1493,15 +1638,12 @@ char *ParseKeyOrValue(char *text,ConfigType type,KeyOrValue *pointer)
           path=URLEncodePath(temppath);
           free(temppath);
 
-          pointer->urlspec=(UrlSpec*)realloc((void*)pointer->urlspec,
-                                             sizeof(UrlSpec)+
-                                             (pointer->urlspec->proto?1+strlen(UrlSpecProto(pointer->urlspec)):0)+
-                                             (pointer->urlspec->host ?1+strlen(UrlSpecHost (pointer->urlspec)):0)+
-                                             strlen(path)+1);
+          pointer->urlspec->path=(unsigned short)(sizeof(UrlSpec)+
+                                                  (pointer->urlspec->proto?1+strlen(UrlSpecProto(pointer->urlspec)):0)+
+                                                  (pointer->urlspec->host ?1+strlen(UrlSpecHost (pointer->urlspec)):0));
 
-          pointer->urlspec->path=sizeof(UrlSpec)+
-                                 (pointer->urlspec->proto?1+strlen(UrlSpecProto(pointer->urlspec)):0)+
-                                 (pointer->urlspec->host ?1+strlen(UrlSpecHost (pointer->urlspec)):0);
+          pointer->urlspec=(UrlSpec*)realloc((void*)pointer->urlspec,
+                                             pointer->urlspec->path+strlen(path)+1);
 
           strcpy(UrlSpecPath(pointer->urlspec),path);
 
@@ -1516,37 +1658,42 @@ char *ParseKeyOrValue(char *text,ConfigType type,KeyOrValue *pointer)
           ;
        else if(*text=='?' && !*(text+1))
          {
-          pointer->urlspec=(UrlSpec*)realloc((void*)pointer->urlspec,sizeof(UrlSpec)+
-                                                                    (pointer->urlspec->proto?1+strlen(UrlSpecProto(pointer->urlspec)):0)+
-                                                                    (pointer->urlspec->host ?1+strlen(UrlSpecHost (pointer->urlspec)) :0)+
-                                                                    (pointer->urlspec->path ?1+strlen(UrlSpecPath (pointer->urlspec)) :0)+
-                                                                    1);
-          pointer->urlspec->args=sizeof(UrlSpec)+
-                                 (pointer->urlspec->proto?1+strlen(UrlSpecProto(pointer->urlspec)):0)+
-                                 (pointer->urlspec->host ?1+strlen(UrlSpecHost (pointer->urlspec)):0)+
-                                 (pointer->urlspec->path ?1+strlen(UrlSpecPath (pointer->urlspec)):0);
-          strcpy(UrlSpecArgs(pointer->urlspec),"");
+          p=text+1;
+
+          pointer->urlspec->args=1;
          }
        else if(*text=='?')
          {
-          char *args=URLRecodeFormArgs(text+1);
+          p=text+1;
 
-          pointer->urlspec=(UrlSpec*)realloc((void*)pointer->urlspec,sizeof(UrlSpec)+
-                                                                    (pointer->urlspec->proto?1+strlen(UrlSpecProto(pointer->urlspec)):0)+
-                                                                    (pointer->urlspec->host ?1+strlen(UrlSpecHost (pointer->urlspec)) :0)+
-                                                                    (pointer->urlspec->path ?1+strlen(UrlSpecPath (pointer->urlspec)) :0)+
-                                                                    strlen(args)+1);
-          pointer->urlspec->args=sizeof(UrlSpec)+
-                                 (pointer->urlspec->proto?1+strlen(UrlSpecProto(pointer->urlspec)):0)+
-                                 (pointer->urlspec->host ?1+strlen(UrlSpecHost (pointer->urlspec)):0)+
-                                 (pointer->urlspec->path ?1+strlen(UrlSpecPath (pointer->urlspec)):0);
-          strcpy(UrlSpecArgs(pointer->urlspec),args);
-
-          free(args);
+          pointer->urlspec->args=1;
          }
        else
-         {errmsg=(char*)malloc(64+strlen(orgtext));sprintf(errmsg,"Expecting a URL-SPECIFICATION, got this '%s'.",orgtext); break;}
+         {errmsg=(char*)malloc(64+strlen(orgtext));sprintf(errmsg,"Expecting a URL-SPECIFICATION, got this '%s'.",orgtext);
+          free(pointer->urlspec); break;}
 
+       if(pointer->urlspec->args)
+         {
+          char *args=NULL;
+
+          if(*p)
+             args=URLRecodeFormArgs(p);
+          else
+             args=p;
+
+          pointer->urlspec->args=(unsigned short)(sizeof(UrlSpec)+
+                                                  (pointer->urlspec->proto?1+strlen(UrlSpecProto(pointer->urlspec)):0)+
+                                                  (pointer->urlspec->host ?1+strlen(UrlSpecHost (pointer->urlspec)):0)+
+                                                  (pointer->urlspec->path ?1+strlen(UrlSpecPath (pointer->urlspec)):0));
+
+          pointer->urlspec=(UrlSpec*)realloc((void*)pointer->urlspec,
+                                             pointer->urlspec->args+strlen(args)+1);
+
+          strcpy(UrlSpecArgs(pointer->urlspec),args);
+
+          if(*args)
+             free(args);
+         }
       }
     break;
    }
