@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/src/RCS/sockets.h 2.9 2001/05/19 18:59:24 amb Exp $
+  $Header: /home/amb/wwwoffle/src/RCS/sockets.h 2.11 2003/11/03 16:31:03 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.6d.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.8a.
   Socket function header file.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1996,97,98,99,2000,01 Andrew M. Bishop
+  This file Copyright 1996,97,98,99,2000,01,02,03 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -23,13 +23,11 @@ int OpenClientSocket(char* host, int port);
 int OpenServerSocket(char* host,int port);
 int AcceptConnect(int socket);
 
-int SocketRemoteName(int socket,char **name,char **ipname,int *port);
-int SocketLocalName(int socket,char **name,char **ipname,int *port);
+int SocketRemoteName(int socket,/*@out@*/ /*@null@*/ char **name,/*@out@*/ /*@null@*/ char **ipname,/*@out@*/ int *port);
+int SocketLocalName(int socket,/*@out@*/ /*@null@*/ char **name,/*@out@*/ /*@null@*/ char **ipname,/*@out@*/ int *port);
 
-int CloseSocket(int socket);
-#ifdef __CYGWIN__
-int CloseCygwinSocket(int socket);
-#endif
+int /*@alt void@*/ CloseSocket(int socket);
+int /*@alt void@*/ ShutdownSocket(int socket);
 
 char *GetFQDN(void);
 
