@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/src/RCS/wwwoffle.c 2.49 2002/02/10 10:41:28 amb Exp $
+  $Header: /home/amb/wwwoffle/src/RCS/wwwoffle.c 2.63 2003/08/28 18:05:52 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.7a.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.8.
   A user level program to interact with the server.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1996,97,98,99,2000,01,02 Andrew M. Bishop
+  This file Copyright 1996,97,98,99,2000,01,02,03 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -23,13 +23,15 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 
-#include "version.h"
 #include "wwwoffle.h"
-#include "document.h"
+#include "io.h"
 #include "misc.h"
+#include "errors.h"
 #include "config.h"
 #include "sockets.h"
-#include "errors.h"
+#include "document.h"
+#include "version.h"
+
 
 #ifndef PATH_MAX
 #define PATH_MAX 4096
@@ -77,9 +79,6 @@ static char **url_file_list=NULL;
 /*+ The number of URLs or files. +*/
 static int n_url_file_list=0;
 
-/*+ A file descriptor for the spool directory; not used anywhere, but required for linking with spool.o. +*/
-int fSpoolDir=-1;
-
 
 /*++++++++++++++++++++++++++++++++++++++
   The main program.
@@ -90,7 +89,7 @@ int main(int argc, char** argv)
  int i;
  int recursive_mode=0,depth=0;
  int force=0;
- int stylesheets=0,images=0,frames=0,scripts=0,objects=0;
+ int stylesheets=0,images=0,frames=0,scripts=0,objects=0,nothing=0;
  char *config_file=NULL;
  int exitval=0;
 
@@ -172,32 +171,40 @@ int main(int argc, char** argv)
             {fprintf(stderr,"wwwoffle: Only one '-gS' argument may be given.\n"); exit(1);}
           stylesheets=1;
          }
-       if(strchr(argv[i]+2,'i'))
+       else if(strchr(argv[i]+2,'i'))
          {
           if(images)
             {fprintf(stderr,"wwwoffle: Only one '-gi' argument may be given.\n"); exit(1);}
           images=1;
          }
-       if(strchr(argv[i]+2,'f'))
+       else if(strchr(argv[i]+2,'f'))
          {
           if(frames)
             {fprintf(stderr,"wwwoffle: Only one '-gf' argument may be given.\n"); exit(1);}
           frames=1;
          }
-       if(strchr(argv[i]+2,'s'))
+       else if(strchr(argv[i]+2,'s'))
          {
           if(scripts)
             {fprintf(stderr,"wwwoffle: Only one '-gs' argument may be given.\n"); exit(1);}
           scripts=1;
          }
-       if(strchr(argv[i]+2,'o'))
+       else if(strchr(argv[i]+2,'o'))
          {
           if(objects)
             {fprintf(stderr,"wwwoffle: Only one '-go' argument may be given.\n"); exit(1);}
           objects=1;
          }
-       if(argv[i][2]==0)
-          fprintf(stderr,"wwwoffle: The '-g' option does nothing on its own.\n");
+       else if(argv[i][2]==0)
+         {
+          if(stylesheets || images || frames || scripts || objects)
+            {fprintf(stderr,"wwwoffle: Cannot have '-g' and '-g[Sifso]' together.\n"); exit(1);}
+          else if(nothing)
+            {fprintf(stderr,"wwwoffle: Only one '-g' argument can be given.\n"); exit(1);}
+          nothing=1;
+         }
+       else
+         {fprintf(stderr,"wwwoffle: The '-g' option does allow '%c' following it.\n",argv[i][2]); exit(1);}
 
        argv[i]=NULL;
        continue;
@@ -221,10 +228,15 @@ int main(int argc, char** argv)
 
        if(argv[i][2])
           depth=atoi(&argv[i][2]);
+       else if(i<(argc-1) && atoi(argv[i+1]))
+         {
+          argv[i++]=NULL;
+          depth=atoi(argv[i]);
+         }
        else
           depth=1;
-       if(depth<=0)
-         {fprintf(stderr,"wwwoffle: The '-%c' argument may only be followed by a positive integer.\n",argv[i][1]); exit(1);}
+       if(depth<0)
+         {fprintf(stderr,"wwwoffle: The '-%c' argument may only be followed by non-negative integer.\n",argv[i][1]); exit(1);}
 
        argv[i]=NULL;
        continue;
@@ -355,6 +367,8 @@ int main(int argc, char** argv)
     add_url_file(argv[i]);
    }
 
+ /* Check special conditions depending on the mode */
+
  if(action==None)
     action=Get;
 
@@ -364,6 +378,18 @@ int main(int argc, char** argv)
     usage(0);
    }
 
+ if((action==Output || action==OutputWithHeader) && n_url_file_list!=1)
+   {
+    fprintf(stderr,"wwwoffle: The -o and -O options require exactly one URL.\n\n");
+    usage(0);
+   }
+
+ if(action==Get && n_url_file_list==0)
+   {
+    fprintf(stderr,"wwwoffle: No URLs were specified to request.\n\n");
+    usage(0);
+   }
+
  /* Initialise things. */
 
  if(!config_file && !host && (env=getenv("WWWOFFLE_PROXY")))
@@ -406,10 +432,12 @@ int main(int argc, char** argv)
 
  if(config_file)
    {
-    init_buffer(2);
+    init_io(2);
 
     if(ReadConfigurationFile(2))
        PrintMessage(Fatal,"Error in configuration file '%s'.",config_file);
+
+    finish_io(2);
    }
 
  if(!host)
@@ -426,11 +454,14 @@ int main(int argc, char** argv)
     char *line=NULL;
 
     socket=OpenClientSocket(host,port?port:wwwoffle_port);
-    init_buffer(socket);
 
     if(socket==-1)
        PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host,port?port:wwwoffle_port);
 
+    init_io(socket);
+    configure_io_read(socket,ConfigInteger(SocketTimeout),0,0);
+    configure_io_write(socket,ConfigInteger(SocketTimeout),0,0);
+
     /* Send the message. */
 
     if(ConfigString(PassWord))
@@ -476,6 +507,7 @@ int main(int argc, char** argv)
           exitval=1;
       }
 
+    finish_io(socket);
     CloseSocket(socket);
    }
 
@@ -491,17 +523,33 @@ int main(int argc, char** argv)
        if(strcmp(url_file_list[i],"-") && stat(url_file_list[i],&buf))
          {
           int socket;
-          char buffer[READ_BUFFER_SIZE];
+          char *p,buffer[READ_BUFFER_SIZE];
+
+          for(p=url_file_list[i];*p;p++)
+             if(*p=='#')
+               {
+                *p=0;
+                break;
+               }
+
+          Url=SplitURL(url_file_list[i]);
+
+          if(!Url->Protocol)
+            {
+             PrintMessage(Warning,"Cannot request protocol '%s'.",Url->proto);
+             continue;
+            }
 
           socket=OpenClientSocket(host,port?port:http_port);
-          init_buffer(socket);
 
           if(socket==-1)
              PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host,port?port:http_port);
 
-          Url=SplitURL(url_file_list[i]);
+          init_io(socket);
+          configure_io_read(socket,ConfigInteger(SocketTimeout),0,0);
+          configure_io_write(socket,ConfigInteger(SocketTimeout),0,0);
 
-          if(recursive_mode || depth || force || stylesheets || images || frames || scripts || objects)
+          if(recursive_mode || depth || force || stylesheets || images || frames || scripts || objects || nothing)
             {
              char *limit="";
 
@@ -524,9 +572,7 @@ int main(int argc, char** argv)
              else
                 depth=0;
 
-             refresh=CreateRefreshPath(Url,limit,depth,
-                                       force,
-                                       stylesheets,images,frames,scripts,objects);
+             refresh=CreateRefreshPath(Url,limit,depth,force,stylesheets,images,frames,scripts,objects);
 
              if(*limit)
                 free(limit);
@@ -546,17 +592,16 @@ int main(int argc, char** argv)
              if(refresh)
                 write_formatted(socket,"GET %s HTTP/1.0\r\n"
                                        "Authorization: Basic %s\r\n"
-                                       "Pragma: wwwoffle\r\n"
+                                       "Pragma: wwwoffle-client\r\n"
                                        "Accept: */*\r\n"
                                        "\r\n",
                                 refresh,userpass2);
              else
                 write_formatted(socket,"GET /refresh/?%s HTTP/1.0\r\n"
-                                       "Authorization: Basic %s\r\n"
-                                       "Pragma: wwwoffle\r\n"
+                                       "Pragma: wwwoffle-client\r\n"
                                        "Accept: */*\r\n"
                                        "\r\n",
-                                URLEncodeFormArgs(Url->name),userpass2);
+                                URLEncodeFormArgs(Url->file));
 
              free(userpass1);
              free(userpass2);
@@ -565,13 +610,13 @@ int main(int argc, char** argv)
             {
              if(refresh)
                 write_formatted(socket,"GET %s HTTP/1.0\r\n"
-                                       "Pragma: wwwoffle\r\n"
+                                       "Pragma: wwwoffle-client\r\n"
                                        "Accept: */*\r\n"
                                        "\r\n",
                                 refresh);
              else
                 write_formatted(socket,"GET /refresh/?%s HTTP/1.0\r\n"
-                                       "Pragma: wwwoffle\r\n"
+                                       "Pragma: wwwoffle-client\r\n"
                                        "Accept: */*\r\n"
                                        "\r\n",
                                 URLEncodeFormArgs(Url->name));
@@ -580,6 +625,7 @@ int main(int argc, char** argv)
           while(read_data(socket,buffer,READ_BUFFER_SIZE)>0)
              ;
 
+          finish_io(socket);
           CloseSocket(socket);
 
           free(refresh);
@@ -594,7 +640,6 @@ int main(int argc, char** argv)
           if(strcmp(url_file_list[i],"-"))
             {
              file=open(url_file_list[i],O_RDONLY);
-             init_buffer(file);
 
              if(file==-1)
                {PrintMessage(Warning,"Cannot open file '%s' for reading.",url_file_list[i]);continue;}
@@ -603,12 +648,13 @@ int main(int argc, char** argv)
             }
           else
             {
-             file=fileno(stdin);
-             init_buffer(file);
+             file=0;
 
              printf("Reading: <stdin>\n");
             }
 
+          init_io(file);
+
           strcpy(buffer,"file://localhost");
           if(*url_file_list[i]=='/')
              strcat(buffer,url_file_list[i]);
@@ -650,6 +696,7 @@ int main(int argc, char** argv)
 
           FreeURL(Url);
           free(buffer);
+          finish_io(file);
           if(file!=0)
              close(file);
          }
@@ -660,22 +707,37 @@ int main(int argc, char** argv)
    {
     URL *Url;
     int socket,n,length=0;
-    char buffer[READ_BUFFER_SIZE];
+    char *p,buffer[READ_BUFFER_SIZE];
     char *data=(char*)malloc(READ_BUFFER_SIZE+1);
 
+    for(p=url_file_list[0];*p;p++)
+       if(*p=='#')
+         {
+          *p=0;
+          break;
+         }
+
+    Url=SplitURL(url_file_list[0]);
+
+    if(!Url->Protocol)
+       PrintMessage(Fatal,"Cannot post or put protocol '%s'.",Url->proto);
+
     socket=OpenClientSocket(host,port?port:http_port);
-    init_buffer(socket);
 
     if(socket==-1)
        PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host,port?port:http_port);
 
-    Url=SplitURL(url_file_list[0]);
+    init_io(socket);
+    configure_io_read(socket,ConfigInteger(SocketTimeout),0,0);
+    configure_io_write(socket,ConfigInteger(SocketTimeout),0,0);
 
     if(action==Post)
        printf("Posting: %s\n",Url->name);
     else
        printf("Putting: %s\n",Url->name);
 
+    init_io(0);
+
     while((n=read_data(0,data+length,READ_BUFFER_SIZE))>0)
       {
        length+=n;
@@ -701,7 +763,7 @@ int main(int argc, char** argv)
                                  "Authorization: Basic %s\r\n"
                                  "Content-Type: application/x-www-form-urlencoded\r\n"
                                  "Content-Length: %d\r\n"
-                                 "Pragma: wwwoffle\r\n"
+                                 "Pragma: wwwoffle-client\r\n"
                                  "Accept: */*\r\n"
                                  "\r\n",
                           Url->name,userpass2,length);
@@ -709,7 +771,7 @@ int main(int argc, char** argv)
           write_formatted(socket,"PUT %s HTTP/1.0\r\n"
                                  "Authorization: Basic %s\r\n"
                                  "Content-Length: %d\r\n"
-                                 "Pragma: wwwoffle\r\n"
+                                 "Pragma: wwwoffle-client\r\n"
                                  "Accept: */*\r\n"
                                  "\r\n",
                           Url->name,userpass2,length);
@@ -723,14 +785,14 @@ int main(int argc, char** argv)
           write_formatted(socket,"POST %s HTTP/1.0\r\n"
                                  "Content-Type: application/x-www-form-urlencoded\r\n"
                                  "Content-Length: %d\r\n"
-                                 "Pragma: wwwoffle\r\n"
+                                 "Pragma: wwwoffle-client\r\n"
                                  "Accept: */*\r\n"
                                  "\r\n",
                           Url->name,length);
        else
           write_formatted(socket,"PUT %s HTTP/1.0\r\n"
                                  "Content-Length: %d\r\n"
-                                 "Pragma: wwwoffle\r\n"
+                                 "Pragma: wwwoffle-client\r\n"
                                  "Accept: */*\r\n"
                                  "\r\n",
                           Url->name,length);
@@ -743,6 +805,7 @@ int main(int argc, char** argv)
     while(read_data(socket,buffer,READ_BUFFER_SIZE)>0)
        ;
 
+    finish_io(socket);
     CloseSocket(socket);
 
     FreeURL(Url);
@@ -752,19 +815,30 @@ int main(int argc, char** argv)
    {
     URL *Url;
     int socket;
-    char *line=NULL,buffer[READ_BUFFER_SIZE];
+    char *line=NULL;
+    char *p,buffer[READ_BUFFER_SIZE];
     int nbytes;
 
-    if(!n_url_file_list)
-       PrintMessage(Fatal,"No URL specified to output.");
+    for(p=url_file_list[0];*p;p++)
+       if(*p=='#')
+         {
+          *p=0;
+          break;
+         }
+
+    Url=SplitURL(url_file_list[0]);
+
+    if(!Url->Protocol)
+       PrintMessage(Fatal,"Cannot fetch data from protocol '%s'.",Url->proto);
 
     socket=OpenClientSocket(host,port?port:http_port);
-    init_buffer(socket);
 
     if(socket==-1)
        PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host,port?port:http_port);
 
-    Url=SplitURL(url_file_list[0]);
+    init_io(socket);
+    configure_io_read(socket,ConfigInteger(SocketTimeout),0,0);
+    configure_io_write(socket,ConfigInteger(SocketTimeout),0,0);
 
     if(action!=Output && action!=OutputWithHeader)
        fprintf(stderr,"Getting: %s\n",Url->name);
@@ -778,7 +852,7 @@ int main(int argc, char** argv)
 
        write_formatted(socket,"GET %s HTTP/1.0\r\n"
                               "Authorization: Basic %s\r\n"
-                              "Pragma: wwwoffle\r\n"
+                              "Pragma: wwwoffle-client\r\n"
                               "Accept: */*\r\n"
                               "\r\n",
                        Url->name,userpass2);
@@ -788,7 +862,7 @@ int main(int argc, char** argv)
       }
     else
        write_formatted(socket,"GET %s HTTP/1.0\r\n"
-                              "Pragma: wwwoffle\r\n"
+                              "Pragma: wwwoffle-client\r\n"
                               "Accept: */*\r\n"
                               "\r\n",
                        Url->name);
@@ -835,6 +909,7 @@ int main(int argc, char** argv)
     else
        PrintMessage(Fatal,"Cannot read from wwwoffle server.");
 
+    finish_io(socket);
     CloseSocket(socket);
 
     FreeURL(Url);
@@ -878,9 +953,9 @@ static void usage(int verbose)
          "       wwwoffle -online | -autodial | -offline | -fetch\n"
          "       wwwoffle -config | -purge | -status | -kill\n"
          "       wwwoffle [-o|-O] <url>\n"
+         "       wwwoffle [-post|-put] <url>\n"
          "       wwwoffle [-g[Sisfo]] [-F] [-(d|r|R)[<depth>]] <url> ...\n"
          "       wwwoffle [-g[Sisfo]] [-F] [-(d|r|R)[<depth>]] [<file>|-] ...\n"
-         "       wwwoffle -post <url> | -put <url>\n"
          "\n"
          "Any of these can also take:  [-p <host>[:<port>] | -c <config-file>]\n"
          "The environment variable WWWOFFLE_PROXY can be set instead of -p or -c options.\n"
@@ -919,8 +994,9 @@ static void usage(int verbose)
             " -o                  : Fetch the URL and output it on the standard output.\n"
             " -O                  : As above but include the HTTP header.\n"
             "\n"
-            " -g[Sisfo]           : Fetch the items included in the specified URLs.\n"
+            " -g[Sisfo]           : Fetch the items included in the specified URLs:\n"
             "                       (S=stylesheets, i=images, f=frames, s=scripts, o=objects)\n"
+            "                       When used without any options fetches none of these.\n"
             " -F                  : Force the url to be refreshed even if already cached.\n"
             " -(d|r|R)[<depth>]   : Fetch pages linked to the URLs and their links,\n"
             "                       going no more than <depth> steps (default 1).\n"
@@ -928,6 +1004,10 @@ static void usage(int verbose)
             "                        (-r => URLs on the same host)\n"
             "                        (-R => URLs on any host)\n"
             "\n"
+            "                      (If the -F, -(d|r|R) or -g[Sisfo] options are set they\n"
+            "                       override the options in the FetchOptions section of the\n"
+            "                       config file and only the -g[Sisfo] options are fetched.)\n"
+            "\n"
             "wwwoffle -post <url> : Create a request using the POST method, the data is read\n"
             "                       from stdin and appended to the request.  The user should\n"
             "                       ensure that the data is correctly url-encoded.\n"
