@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/src/RCS/wwwoffled.c 2.43 2002/03/03 19:27:46 amb Exp $
+  $Header: /home/amb/wwwoffle/src/RCS/wwwoffled.c 2.66 2003/10/19 15:40:45 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.7a.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.8.
   A demon program to maintain the database and spawn the servers.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1996,97,98,99,2000,01,02 Andrew M. Bishop
+  This file Copyright 1996,97,98,99,2000,01,02,03 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -41,18 +41,14 @@
 #include <signal.h>
 #include <grp.h>
 
-#include "version.h"
 #include "wwwoffle.h"
+#include "io.h"
 #include "misc.h"
+#include "errors.h"
 #include "config.h"
 #include "sockets.h"
-#include "errors.h"
+#include "version.h"
 
-#if defined (hpux)
-/* HP/UX has no `seteuid'. Instead it uses `setresuid' to set `uid', `euid', and `suid' simultaneously. */
-#define setegid(egid)   setresgid (-1, (egid), -1)
-#define seteuid(euid)   setresuid (-1, (euid), -1)
-#endif
 
 static void usage(int verbose);
 static void demoninit(void);
@@ -95,11 +91,8 @@ int purge_pid=0;
 /*+ The current status, fetching or not. +*/
 int fetching=0;
 
-/*+ The file descriptor of the sppol directory. +*/
-int fSpoolDir=-1;
-
 /*+ Set to 1 when the demon is to shutdown. +*/
-static int closedown=0;
+int closedown=0;
 
 /*+ Set to 1 when the demon is to re-read config file due to a sighup. +*/
 static int readconfig=0;
@@ -116,6 +109,9 @@ static int print_pid=0;
 /*+ True if the -f option was passed on the command line. +*/
 int nofork=0;
 
+/* Global variables used to pass on information about remote client to CGI. */
+extern char *client_hostname,*client_ip;
+
 
 /*++++++++++++++++++++++++++++++++++++++
   The main program.
@@ -127,6 +123,7 @@ int main(int argc, char** argv)
  int err;
  struct stat buf;
  char *config_file=NULL;
+ int uid,gid;
 
  /* Parse the command line options */
 
@@ -141,12 +138,15 @@ int main(int argc, char** argv)
     if(!strcmp(argv[i],"-d"))
       {
        detached=0;
-       if(i<(argc-1) && isdigit(argv[i+1][0]) && !argv[i+1][1])
+
+       if(i<(argc-1) && isdigit(argv[i+1][0]))
          {
-          DebugLevel=Fatal+1-atoi(argv[++i]);
-          if(DebugLevel<0)
-             DebugLevel=0;
+          DebuggingLevel=Fatal+1-atoi(argv[++i]);
+
+          if(DebuggingLevel<0 || DebuggingLevel>Fatal)
+            {fprintf(stderr,"wwwoffled: The '-d' option requires a number between 1 and %d.\n",Fatal+1); exit(1);}
          }
+
        continue;
       }
 
@@ -189,9 +189,12 @@ int main(int argc, char** argv)
 
  InitConfigurationFile(config_file);
 
- init_buffer(2);
+ init_io(2);
+
  if(ReadConfigurationFile(2))
-    PrintMessage(Fatal,"Error in configuration file '%s'.",config_file);
+    PrintMessage(Fatal,"Error in configuration file '%s'.",ConfigurationFileName());
+
+ finish_io(2);
 
  InitErrorHandler("wwwoffled",ConfigInteger(UseSyslog),1); /* enable syslog if requested. */
 
@@ -210,38 +213,79 @@ int main(int argc, char** argv)
 #endif
 
  PrintMessage(Important,"WWWOFFLE Demon Version %s (%s,%s) started.",WWWOFFLE_VERSION,ZLIB_STRING,IPV6_STRING);
- PrintMessage(Inform,"WWWOFFLE Read Configuration File '%s'.",config_file);
-
+ PrintMessage(Inform,"WWWOFFLE Read Configuration File '%s'.",ConfigurationFileName());
 
  /* Change the user and group. */
 
- if(ConfigInteger(WWWOFFLE_Gid) != -1)
+ gid=ConfigInteger(WWWOFFLE_Gid);
+ uid=ConfigInteger(WWWOFFLE_Uid);
+
+ if(uid!=-1)
+    seteuid(0);
+
+ if(gid!=-1)
    {
-#if !defined(__CYGWIN__)
-    if(getuid()==0)
+#if HAVE_SETGROUPS
+    if(getuid()==0 || geteuid()==0)
        if(setgroups(0, NULL)<0)
           PrintMessage(Fatal,"Cannot clear supplementary group list [%!s].");
 #endif
 
-    if(setgid(ConfigInteger(WWWOFFLE_Gid))<0)
-       PrintMessage(Fatal,"Cannot set group id to %d [%!s].",ConfigInteger(WWWOFFLE_Gid));
-    if(setegid(ConfigInteger(WWWOFFLE_Gid))<0)
-       PrintMessage(Fatal,"Cannot set effective group id to %d [%!s].",ConfigInteger(WWWOFFLE_Gid));
+#if HAVE_SETRESGID
+    if(setresgid(gid,gid,gid)<0)
+       PrintMessage(Fatal,"Cannot set real/effective/saved group id to %d [%!s].",gid);
+#else
+    if(geteuid()==0)
+      {
+       if(setgid(gid)<0)
+          PrintMessage(Fatal,"Cannot set group id to %d [%!s].",gid);
+      }
+    else
+      {
+#if HAVE_SETREGID
+       if(setregid(getegid(),gid)<0)
+          PrintMessage(Fatal,"Cannot set effective group id to %d [%!s].",gid);
+       if(setregid(gid,-1)<0)
+          PrintMessage(Fatal,"Cannot set real group id to %d [%!s].",gid);
+#else
+       PrintMessage(Fatal,"Must be root to totally change group id.");
+#endif
+      }
+#endif
    }
 
- if(ConfigInteger(WWWOFFLE_Uid) != -1)
+ if(uid!=-1)
    {
-    if(setuid(ConfigInteger(WWWOFFLE_Uid))<0)
-       PrintMessage(Fatal,"Cannot set user id to %d [%!s].",ConfigInteger(WWWOFFLE_Uid));
-    if(seteuid(ConfigInteger(WWWOFFLE_Uid))<0)
-       PrintMessage(Fatal,"Cannot set effective user id to %d [%!s].",ConfigInteger(WWWOFFLE_Uid));
+#if HAVE_SETRESUID
+    if(setresuid(uid,uid,uid)<0)
+       PrintMessage(Fatal,"Cannot set real/effective/saved user id to %d [%!s].",uid);
+#else
+    if(geteuid()==0)
+      {
+       if(setuid(uid)<0)
+          PrintMessage(Fatal,"Cannot set user id to %d [%!s].",uid);
+      }
+    else
+      {
+#if HAVE_SETREUID
+       if(setreuid(geteuid(),uid)<0)
+          PrintMessage(Fatal,"Cannot set effective user id to %d [%!s].",uid);
+       if(setreuid(uid,-1)<0)
+          PrintMessage(Fatal,"Cannot set real user id to %d [%!s].",uid);
+#else
+       PrintMessage(Fatal,"Must be root to totally change user id.");
+#endif
+      }
+#endif
    }
 
- if(ConfigInteger(WWWOFFLE_Gid) != -1 || ConfigInteger(WWWOFFLE_Uid) != -1)
-    PrintMessage(Inform,"Running with uid=%d, gid=%d.",getuid(),getgid());
+ if(uid!=-1 || gid!=-1)
+    PrintMessage(Inform,"Running with uid=%d, gid=%d.",geteuid(),getegid());
 
+ if(geteuid()==0 || getegid()==0)
+    PrintMessage(Warning,"Running with root user or group privileges is not recommended.");
 
- /* Bind the HTTP proxy sockets. */
+ /* Bind the HTTP proxy socket(s). */
 
 #if USE_IPV6
  if(ConfigString(Bind_IPv6))
@@ -252,40 +296,33 @@ int main(int argc, char** argv)
    }
 #endif
 
-#if USE_IPV6
- if(http_fd[1]!=-1 && ConfigString(Bind_IPv4) && ConfigString(Bind_IPv6) &&
-    !strcmp(ConfigString(Bind_IPv6),"0:0:0:0:0:0:0:0") && !strcmp(ConfigString(Bind_IPv4),"0.0.0.0"))
-    PrintMessage(Inform,"Not creating HTTP IPv4 server socket on all addresses since IPv6 one is bound to all addresses.");
- else
+ if(ConfigString(Bind_IPv4))
    {
-#endif
-    if(ConfigString(Bind_IPv4))
+    http_fd[0]=OpenServerSocket(ConfigString(Bind_IPv4),ConfigInteger(HTTP_Port));
+    if(http_fd[0]==-1)
       {
-       http_fd[0]=OpenServerSocket(ConfigString(Bind_IPv4),ConfigInteger(HTTP_Port));
-       if(http_fd[0]==-1)
-         {
 #if USE_IPV6
+       if(http_fd[1]!=-1 && ConfigString(Bind_IPv4) && !strcmp(ConfigString(Bind_IPv4),"0.0.0.0") &&
+                            ConfigString(Bind_IPv6) && !strcmp(ConfigString(Bind_IPv6),"0:0:0:0:0:0:0:0"))
+          PrintMessage(Warning,"Cannot create HTTP IPv4 server socket (but the IPv6 one might accept IPv4 connections).");
+       else
           PrintMessage(Fatal,"Cannot create HTTP IPv4 server socket.");
 #else
-          PrintMessage(Fatal,"Cannot create HTTP server socket.");
+       PrintMessage(Fatal,"Cannot create HTTP server socket.");
 #endif
-         }
       }
-#if USE_IPV6
    }
-#endif
 
  if(http_fd[0]==-1 && http_fd[1]==-1)
    {
 #if USE_IPV6
-    PrintMessage(Fatal,"Neither IPv4 or IPv6 HTTP socket could be bound.");
+    PrintMessage(Fatal,"The IPv4 and IPv6 HTTP sockets were not bound; are they disabled in the config file?");
 #else
-    PrintMessage(Fatal,"The HTTP socket could not be bound.");
+    PrintMessage(Fatal,"The HTTP socket was not bound; are they disabled in the config file?");
 #endif
    }
 
-
- /* Bind the WWWOFFLE control sockets. */
+ /* Bind the WWWOFFLE control socket(s). */
 
 #if USE_IPV6
  if(ConfigString(Bind_IPv6))
@@ -296,35 +333,29 @@ int main(int argc, char** argv)
    }
 #endif
 
-#if USE_IPV6
- if(wwwoffle_fd[1]!=-1 && ConfigString(Bind_IPv4) && ConfigString(Bind_IPv6) &&
-    !strcmp(ConfigString(Bind_IPv6),"0:0:0:0:0:0:0:0") && !strcmp(ConfigString(Bind_IPv4),"0.0.0.0"))
-    PrintMessage(Inform,"Not creating WWWOFFLE IPv4 server socket on all addresses since IPv6 one is bound to all addresses.");
- else
+ if(ConfigString(Bind_IPv4))
    {
-#endif
-    if(ConfigString(Bind_IPv4))
+    wwwoffle_fd[0]=OpenServerSocket(ConfigString(Bind_IPv4),ConfigInteger(WWWOFFLE_Port));
+    if(wwwoffle_fd[0]==-1)
       {
-       wwwoffle_fd[0]=OpenServerSocket(ConfigString(Bind_IPv4),ConfigInteger(WWWOFFLE_Port));
-       if(wwwoffle_fd[0]==-1)
-         {
 #if USE_IPV6
+       if(wwwoffle_fd[1]!=-1 && ConfigString(Bind_IPv4) && !strcmp(ConfigString(Bind_IPv4),"0.0.0.0") &&
+                                ConfigString(Bind_IPv6) && !strcmp(ConfigString(Bind_IPv6),"0:0:0:0:0:0:0:0"))
+          PrintMessage(Warning,"Cannot create WWWOFFLE IPv4 server socket (but the IPv6 one might accept IPv4 connections).");
+       else
           PrintMessage(Fatal,"Cannot create WWWOFFLE IPv4 server socket.");
 #else
-          PrintMessage(Fatal,"Cannot create WWWOFFLE server socket.");
+       PrintMessage(Fatal,"Cannot create WWWOFFLE server socket.");
 #endif
-         }
       }
-#if USE_IPV6
    }
-#endif
 
  if(wwwoffle_fd[0]==-1 && wwwoffle_fd[1]==-1)
    {
 #if USE_IPV6
-    PrintMessage(Fatal,"Neither IPv4 or IPv6 WWWOFFLE socket could be bound.");
+    PrintMessage(Fatal,"The IPv4 and IPv6 WWWOFFLE sockets were not bound; are they disabled in the config file?");
 #else
-    PrintMessage(Fatal,"The WWWOFFLE socket could not be bound.");
+    PrintMessage(Fatal,"The WWWOFFLE socket was not bound; are they disabled in the config file?");
 #endif
    }
 
@@ -343,16 +374,10 @@ int main(int argc, char** argv)
  if(!S_ISDIR(buf.st_mode))
     PrintMessage(Fatal,"The spool directory %s is not a directory.",SpoolDir);
 
- err=chdir(ConfigString(SpoolDir));
+ err=ChangeToSpoolDir(ConfigString(SpoolDir));
  if(err==-1)
     PrintMessage(Fatal,"Cannot change to spool directory %s [%!s].",ConfigString(SpoolDir));
 
-#if !defined(__CYGWIN__)
- fSpoolDir=open(".",O_RDONLY);
- if(fSpoolDir==-1)
-    PrintMessage(Fatal,"Cannot open the spool directory '%s' [%!s].",ConfigString(SpoolDir));
-#endif
-
  /* Detach from terminal */
 
  if(detached)
@@ -411,23 +436,43 @@ int main(int argc, char** argv)
              int port,client;
 
              client=AcceptConnect(wwwoffle_fd[nfd]);
-             init_buffer(client);
 
-             if(client>=0 && !SocketRemoteName(client,&host,&ip,&port))
+             if(client>=0)
                {
-                if(IsAllowedConnectHost(host) || IsAllowedConnectHost(CanonicaliseHost(ip)))
+                init_io(client);
+                configure_io_read(client,ConfigInteger(SocketTimeout),0,0);
+                configure_io_write(client,ConfigInteger(SocketTimeout),0,0);
+
+                if(SocketRemoteName(client,&host,&ip,&port))
                   {
-                   PrintMessage(Important,"WWWOFFLE Connection from host %s (%s).",host,ip); /* Used in audit-usage.pl */
-
-                   CommandConnect(client);
-
-                   if(fetch_fd!=client)
-                      CloseSocket(client);
+                   finish_io(client);
+                   CloseSocket(client);
                   }
                 else
                   {
-                   PrintMessage(Warning,"WWWOFFLE Connection rejected from host %s (%s).",host,ip); /* Used in audit-usage.pl */
-                   CloseSocket(client);
+                   char *canonical_ip=CanonicaliseHost(ip);
+
+                   if(IsAllowedConnectHost(host) || IsAllowedConnectHost(canonical_ip))
+                     {
+                      PrintMessage(Important,"WWWOFFLE Connection from host %s (%s).",host,canonical_ip); /* Used in audit-usage.pl */
+
+                      CommandConnect(client);
+
+                      if(fetch_fd!=client)
+                        {
+                         finish_io(client);
+                         CloseSocket(client);
+                        }
+                     }
+                   else
+                     {
+                      PrintMessage(Warning,"WWWOFFLE Connection rejected from host %s (%s).",host,canonical_ip); /* Used in audit-usage.pl */
+                      finish_io(client);
+                      CloseSocket(client);
+                     }
+
+                   if(canonical_ip!=ip)
+                      free(canonical_ip);
                   }
                }
             }
@@ -438,22 +483,41 @@ int main(int argc, char** argv)
              int port,client;
 
              client=AcceptConnect(http_fd[nfd]);
-             init_buffer(client);
 
-             if(client>=0 && !SocketRemoteName(client,&host,&ip,&port))
+             if(client>=0)
                {
-                if(IsAllowedConnectHost(host) || IsAllowedConnectHost(CanonicaliseHost(ip)))
+                init_io(client);
+                configure_io_read(client,ConfigInteger(SocketTimeout),0,0);
+                configure_io_write(client,ConfigInteger(SocketTimeout),0,0);
+
+                if(!SocketRemoteName(client,&host,&ip,&port))
                   {
-                   PrintMessage(Inform,"HTTP Proxy connection from host %s (%s).",host,ip); /* Used in audit-usage.pl */
-                   ForkServer(client,1);
+                   char *canonical_ip=CanonicaliseHost(ip);
+
+                   if(IsAllowedConnectHost(host) || IsAllowedConnectHost(canonical_ip))
+                     {
+                      PrintMessage(Inform,"HTTP Proxy connection from host %s (%s).",host,canonical_ip); /* Used in audit-usage.pl */
+
+                      client_hostname=host;
+                      client_ip=canonical_ip;
+
+                      ForkServer(client);
+                     }
+                   else
+                      PrintMessage(Warning,"HTTP Proxy connection rejected from host %s (%s).",host,canonical_ip); /* Used in audit-usage.pl */
+
+                   if(canonical_ip!=ip)
+                      free(canonical_ip);
                   }
-                else
-                   PrintMessage(Warning,"HTTP Proxy connection rejected from host %s (%s).",host,ip); /* Used in audit-usage.pl */
 
                 if(nofork)
                    closedown=1;
 
-                CloseSocket(client);
+                if(!nofork)
+                  {
+                   finish_io(client);
+                   CloseSocket(client);
+                  }
                }
             }
 #if USE_IPV6
@@ -481,6 +545,7 @@ int main(int argc, char** argv)
        int isserver=0;
 
        /* To avoid race conditions, reset the flag before fetching the status */
+
        got_sigchld=0;
 
        while((pid=waitpid(-1,&status,WNOHANG))>0)
@@ -520,6 +585,7 @@ int main(int argc, char** argv)
                }
 
           /* Check if the child that terminated is one of the fetching wwwoffles */
+
           for(i=0;i<max_fetch_servers;i++)
              if(fetch_pids[i]==pid)
                {
@@ -546,30 +612,32 @@ int main(int argc, char** argv)
        start fetch servers to look for jobs in the spool directory. */
 
     while(fetching && n_fetch_servers<max_fetch_servers && n_servers<max_servers)
-       ForkServer(fetch_fd,0);
+       ForkServer(fetch_fd);
 
     if(fetch_fd!=-1 && !fetching && n_fetch_servers==0)
       {
        write_string(fetch_fd,"WWWOFFLE No more to fetch.\n");
+
+       finish_io(fetch_fd);
        CloseSocket(fetch_fd);
        fetch_fd=-1;
 
        PrintMessage(Important,"WWWOFFLE Fetch finished.");
 
-       ForkRunModeScript(ConfigString(RunFetch),"fetch","stop",fetch_fd);
+       ForkRunModeScript(ConfigString(RunFetch),"fetch","stop",-1);
       }
    }
  while(!closedown);
 
  /* Close down and exit. */
 
- if(http_fd[0]!=-1) CloseSocket(http_fd[0]);
- if(wwwoffle_fd[0]!=-1) CloseSocket(wwwoffle_fd[0]);
+ /* These four sockets don't need finish_io() calling because they never
+    had init_io() called, they are just bound to a port listening. */
 
-#if USE_IPV6
+ if(http_fd[0]!=-1) CloseSocket(http_fd[0]);
  if(http_fd[1]!=-1) CloseSocket(http_fd[1]);
+ if(wwwoffle_fd[0]!=-1) CloseSocket(wwwoffle_fd[0]);
  if(wwwoffle_fd[1]!=-1) CloseSocket(wwwoffle_fd[1]);
-#endif
 
  if(!nofork)
    {
@@ -622,10 +690,6 @@ int main(int argc, char** argv)
 
  FinishConfigurationFile();
 
-#if !defined(__CYGWIN__)
- close(fSpoolDir);
-#endif
-
  return(0);
 }
 
@@ -653,8 +717,8 @@ static void usage(int verbose)
 
  if(verbose)
     fprintf(stderr,
-            "(c) Andrew M. Bishop 1996,97,98,99,2000,01 [       amb@gedanken.demon.co.uk ]\n"
-            "                                           [http://www.gedanken.demon.co.uk/]\n"
+            "(c) Andrew M. Bishop 1996,97,98,99,2000,01,02,03 [       amb@gedanken.demon.co.uk ]\n"
+            "                                                 [http://www.gedanken.demon.co.uk/]\n"
             "\n");
 
  fprintf(stderr,
@@ -770,7 +834,7 @@ static void install_sighandlers(void)
   int signum The signal number.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static void sigchild(int signum)
+static void sigchild(/*@unused@*/ int signum)
 {
  got_sigchld=1;
 }
@@ -782,7 +846,7 @@ static void sigchild(int signum)
   int signum The signal number.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static void sigexit(int signum)
+static void sigexit(/*@unused@*/ int signum)
 {
  closedown=1;
 }
@@ -794,7 +858,7 @@ static void sigexit(int signum)
   int signum The signal number.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static void sighup(int signum)
+static void sighup(/*@unused@*/ int signum)
 {
  readconfig=1;
 }
